# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def flatten(array, axis=1, highlevel=True, behavior=None):

    """
    Args:
        array: Data containing nested lists to flatten.
        axis (None or int): If None, the operation flattens all levels of
            nesting, returning a 1-dimensional array. Otherwise, it flattens
            at a specified depth. The outermost dimension is `0`, followed
            by `1`, etc., and negative values count backward from the
            innermost: `-1` is the innermost dimension, `-2` is the next
            level up, etc.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array with one level of nesting removed by erasing the
    boundaries between consecutive lists. Since this operates on a level of
    nesting, `axis=0` is a special case that only removes values at the
    top level that are equal to None.

    Consider the following doubly nested `array`.

        ak.Array([[
                   [1.1, 2.2, 3.3],
                   [],
                   [4.4, 5.5],
                   [6.6]],
                  [],
                  [
                   [7.7],
                   [8.8, 9.9]
                  ]])

    At `axis=1`, the outer lists (length 4, length 0, length 2) become a single
    list (of length 6).

        >>> print(ak.flatten(array, axis=1))
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7], [8.8, 9.9]]

    At `axis=2`, the inner lists (lengths 3, 0, 2, 1, 1, and 2) become three
    lists (of lengths 6, 0, and 3).

        >>> print(ak.flatten(array, axis=2))
        [[1.1, 2.2, 3.3, 4.4, 5.5, 6.6], [], [7.7, 8.8, 9.9]]

    There's also an option to completely flatten the array with `axis=None`.
    This is useful for passing the data to a function that doesn't care about
    nested structure, such as a plotting routine.

        >>> print(ak.flatten(array, axis=None))
        [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    Missing values are eliminated by flattening: there is no distinction
    between an empty list and a value of None at the level of flattening.

        >>> array = ak.Array([[1.1, 2.2, 3.3], None, [4.4], [], [5.5]])
        >>> ak.flatten(array, axis=1)
        <Array [1.1, 2.2, 3.3, 4.4, 5.5] type='5 * float64'>

    As a consequence, flattening at `axis=0` does only one thing: it removes
    None values from the top level.

        >>> ak.flatten(array, axis=0)
        <Array [[1.1, 2.2, 3.3], [4.4], [], [5.5]] type='4 * var * float64'>

    As a technical detail, the flattening operation can be trivial in a common
    case, #ak.layout.ListOffsetArray in which the first `offset` is `0`.
    In that case, the flattened data is simply the array node's `content`.

        >>> array.layout
        <ListOffsetArray64>
            <offsets><Index64 i="[0 4 4 6]" offset="0" length="4"/></offsets>
            <content><ListOffsetArray64>
                <offsets><Index64 i="[0 3 3 5 6 7 9]" offset="0" length="7"/></offsets>
                <content>
                    <NumpyArray format="d" shape="9" data="1.1 2.2 3.3 4.4 5.5 6.6 7.7 8.8 9.9"/>
                </content>
            </ListOffsetArray64></content>
        </ListOffsetArray64>
        >>> np.asarray(array.layout.content.content)
        array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])

    However, it is important to keep in mind that this is a special case:
    #ak.flatten and `content` are not interchangeable!
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    nplike = ak.nplike.of(layout)

    if axis is None:
        out = layout.completely_flatten(function_name="ak.flatten")
        assert isinstance(out, tuple) and all(isinstance(x, np.ndarray) for x in out)

        out = ak._v2.contents.NumpyArray(nplike.concatenate(out))

    elif axis == 0 or layout.axis_wrap_if_negative(axis) == 0:

        def apply(layout):
            if layout.is_UnknownType:
                return apply(ak._v2.contents.NumpyArray(nplike.array([])))

            elif layout.is_IndexedType:
                return apply(layout.project())

            elif layout.is_UnionType:
                if not any(
                    x.is_OptionType and not isinstance(x, ak._v2.contents.UnmaskedArray)
                    for x in layout.contents
                ):
                    return layout

                tags = nplike.asarray(layout.tags)
                index = nplike.array(nplike.asarray(layout.index), copy=True)
                bigmask = nplike.empty(len(index), dtype=np.bool_)
                for tag, content in enumerate(layout.contents):
                    if content.is_OptionType and not isinstance(
                        content, ak._v2.contents.UnmaskedArray
                    ):
                        bigmask[:] = False
                        bigmask[tags == tag] = nplike.asarray(content.bytemask()).view(
                            np.bool_
                        )
                        index[bigmask] = -1

                good = index >= 0
                return ak._v2.contents.UnionArray(
                    ak._v2.index.Index8(tags[good]),
                    ak._v2.index.Index64(index[good]),
                    layout.contents,
                )

            elif layout.is_OptionType:
                return layout.project()

            else:
                return layout

        out = apply(layout)

        return ak._v2._util.wrap(out, behavior, highlevel)

    else:
        out = layout.flatten(axis)
    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
from awkward._v2.forms.form import Form, _parameters_equal


class RegularForm(Form):
    is_ListType = True
    is_RegularType = True

    def __init__(
        self, content, size, has_identifier=False, parameters=None, form_key=None
    ):
        if not isinstance(content, Form):
            raise TypeError(
                "{} all 'contents' must be Form subclasses, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not ak._util.isint(size):
            raise TypeError(
                "{} 'size' must be of type int, not {}".format(
                    type(self).__name__, repr(size)
                )
            )

        self._content = content
        self._size = int(size)
        self._init(has_identifier, parameters, form_key)

    @property
    def content(self):
        return self._content

    @property
    def size(self):
        return self._size

    def __repr__(self):
        args = [repr(self._content), repr(self._size)] + self._repr_args()
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _tolist_part(self, verbose, toplevel):
        return self._tolist_extra(
            {
                "class": "RegularArray",
                "size": self._size,
                "content": self._content._tolist_part(verbose, toplevel=False),
            },
            verbose,
        )

    def _type(self, typestrs):
        return ak._v2.types.regulartype.RegularType(
            self._content._type(typestrs),
            self._size,
            self._parameters,
            ak._v2._util.gettypestr(self._parameters, typestrs),
        )

    def __eq__(self, other):
        if isinstance(other, RegularForm):
            return (
                self._has_identifier == other._has_identifier
                and self._form_key == other._form_key
                and self._size == other._size
                and _parameters_equal(self._parameters, other._parameters)
                and self._content == other._content
            )
        else:
            return False

    def generated_compatibility(self, other):
        if other is None:
            return True

        elif isinstance(other, RegularForm):
            return (
                self._size == other._size
                and _parameters_equal(self._parameters, other._parameters)
                and self._content.generated_compatibility(other._content)
            )

        else:
            return False

    def _getitem_range(self):
        return RegularForm(
            self._content._getitem_range(),
            self._size,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def _getitem_field(self, where, only_fields=()):
        return RegularForm(
            self._content._getitem_field(where, only_fields),
            self._size,
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return RegularForm(
            self._content._getitem_fields(where, only_fields),
            self._size,
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _carry(self, allow_lazy):
        return RegularForm(
            self._content._carry(allow_lazy),
            self._size,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def purelist_parameter(self, key):
        if self._parameters is None or key not in self._parameters:
            return self._content.purelist_parameter(key)
        else:
            return self._parameters[key]

    @property
    def purelist_isregular(self):
        return self._content.purelist_isregular

    @property
    def purelist_depth(self):
        if self.parameter("__array__") in ("string", "bytestring"):
            return 1
        else:
            return self._content.purelist_depth + 1

    @property
    def minmax_depth(self):
        if self.parameter("__array__") in ("string", "bytestring"):
            return (1, 1)
        else:
            mindepth, maxdepth = self._content.minmax_depth
            return (mindepth + 1, maxdepth + 1)

    @property
    def branch_depth(self):
        if self.parameter("__array__") in ("string", "bytestring"):
            return (False, 1)
        else:
            branch, depth = self._content.branch_depth
            return (branch, depth + 1)

    @property
    def fields(self):
        return self._content.fields

    @property
    def dimension_optiontype(self):
        return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import json
import os

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


def from_json_schema(
    source,
    schema,
    highlevel=True,
    behavior=None,
    output_initial_size=1024,
    output_resize_factor=1.5,
):
    """
    Args:
        source (str or bytes): JSON-formatted string to convert into an array.
        schema (str, bytes, or nested dicts): JSONSchema to assume in the parsing.
            The JSON data are *not* validated against the schema; the schema is
            only used to accelerate parsing.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        output_initial_size (int): Initial size (in bytes) of output buffers, which
            grow as needed to accommodate the size of the dataset.
        output_resize_factor (float): Resize multiplier for output buffers, which
            determines how quickly they grow; should be strictly greater than 1.

    Converts a JSON string into an Awkward Array, using a JSONSchema to accelerate
    the parsing of the source and building of the output. The JSON data are not
    *validated* against the schema; the schema is *assumed* to be correct.

    Supported JSONSchema elements:

      * The root of the schema must be `"type": "array"` or `"type": "object"`.
      * Every level must have a `"type"`, which can only name one type (as a string
        or length-1 list) or one type and `"null"` (as a length-2 list).
      * `"type": "boolean"` \u2192 1-byte boolean values.
      * `"type": "integer"` \u2192 8-byte integer values.
        Numbers may include a fractional part, as per the JSONSchema specification,
        but this function ignores any fractional part.
      * `"type": "number"` \u2192 8-byte floating-point values.
      * `"type": "string"` \u2192 UTF-8 encoded strings. All JSON escape sequences are
        supported. Remember that the `source` data are ASCII; Unicode is derived from
        "`\\uXXXX`" escape sequences. If an `"enum"` is given, strings are represented
        as categorical values (#ak.layout.IndexedArray or #ak.layout.IndexedOptionArray).
      * `"type": "array"` \u2192 nested lists. The `"items"` must be specified. If
        `"minItems"` and `"maxItems"` are specified and equal to each other, the
        list has regular-type (#ak.types.RegularType); otherwise, it has variable-length
        type (#ak.types.ListType).
      * `"type": "object"` \u2192 nested records. The `"properties"` must be specified,
        and any properties in the data not described by `"properties"` will not
        appear in the output.

    See also #ak.from_json and #ak.to_json.
    """
    if not isinstance(source, bytes) and not ak._v2._util.isstr(source):
        raise NotImplementedError("for now, 'source' must be bytes or str")

    if isinstance(schema, bytes) or ak._v2._util.isstr(schema):
        schema = json.loads(schema)

    if not isinstance(schema, dict):
        raise TypeError(f"malformed JSONSchema: expected dict, got {schema!r}")

    container = {}
    instructions = []

    if schema.get("type") == "array":
        if "items" not in schema:
            raise TypeError("JSONSchema type is not concrete: array without items")

        instructions.append(["TopLevelArray"])
        form = build_assembly(schema["items"], container, instructions)

    elif schema.get("type") == "object":
        form = build_assembly(schema, container, instructions)

    else:
        raise TypeError(
            "only 'array' and 'object' types supported at the JSONSchema root"
        )

    specializedjson = ak._ext.SpecializedJSON(
        json.dumps(instructions),
        output_initial_size=output_initial_size,
        output_resize_factor=output_resize_factor,
    )

    if not specializedjson.parse_string(source):
        position = specializedjson.json_position
        before = source[max(0, position - 30) : position]
        if isinstance(before, bytes):
            before = before.decode("ascii", errors="surrogateescape")
        before = before.replace(os.linesep, repr(os.linesep).strip("'\""))
        if position - 30 > 0:
            before = "..." + before
        after = source[position : position + 30]
        if isinstance(after, bytes):
            after = after.decode("ascii", errors="surrogateescape")
        if position + 30 < len(source):
            after = after + "..."
        raise ValueError(
            "JSON is invalid or does not fit schema at position {}:\n\n    {}\n    {}".format(
                position, before + after, "-" * len(before) + "^"
            )
        )

    for key, value in container.items():
        if value is None:
            container[key] = specializedjson[key]

    if schema.get("type") == "array":
        length = len(specializedjson)
    else:
        length = 1

    out = ak._v2.operations.convert.from_buffers(form, length, container)

    if schema.get("type") == "array":
        return out
    else:
        return out[0]


def build_assembly(schema, container, instructions):
    if not isinstance(schema, dict):
        raise TypeError(f"unrecognized JSONSchema: expected dict, got {schema!r}")

    if "type" not in schema is None:
        raise TypeError(f"unrecognized JSONSchema: no 'type' in {schema!r}")

    tpe = schema["type"]

    is_optional = False
    if isinstance(tpe, list):
        if "null" in tpe:
            is_optional = True
            tpe = [x for x in tpe if x != "null"]
        if len(tpe) == 1:
            tpe = tpe[0]

    if tpe in {"boolean", "integer", "number"}:
        # https://json-schema.org/understanding-json-schema/reference/boolean.html
        # https://json-schema.org/understanding-json-schema/reference/numeric.html

        if tpe == "boolean":
            instruction = "FillBoolean"
            dtype = "uint8"
            primitive = "bool"
        elif tpe == "integer":
            instruction = "FillInteger"
            primitive = dtype = "int64"
        elif tpe == "number":
            instruction = "FillNumber"
            primitive = dtype = "float64"

        if is_optional:
            mask = f"node{len(container)}"
            container[mask + "-mask"] = None
            node = f"node{len(container)}"
            container[node + "-data"] = None
            instructions.append(["FillByteMaskedArray", mask + "-mask", "int8"])
            instructions.append([instruction, node + "-data", dtype])
            return ak._v2.forms.ByteMaskedForm(
                "i8",
                ak._v2.forms.NumpyForm(primitive, form_key=node),
                valid_when=True,
                form_key=mask,
            )

        else:
            node = f"node{len(container)}"
            container[node + "-data"] = None
            instructions.append([instruction, node + "-data", dtype])
            return ak._v2.forms.NumpyForm(primitive, form_key=node)

    elif tpe == "string":
        # https://json-schema.org/understanding-json-schema/reference/string.html#string
        if "enum" in schema:
            strings = schema["enum"]
            assert isinstance(strings, list)
            assert len(strings) >= 1
            assert all(ak._v2._util.isstr(x) for x in strings)
            bytestrings = [x.encode("utf-8", errors="surrogateescape") for x in strings]

            index = f"node{len(container)}"
            container[index + "-index"] = None
            offsets = f"node{len(container)}"
            container[offsets + "-offsets"] = numpy.empty(len(strings) + 1, np.int64)
            container[offsets + "-offsets"][0] = 0
            container[offsets + "-offsets"][1:] = numpy.cumsum(
                [len(x) for x in bytestrings]
            )
            node = f"container{len(container)}"
            container[node + "-data"] = b"".join(bytestrings)

            if is_optional:
                instruction = "FillNullEnumString"
                formtype = ak._v2.forms.IndexedOptionForm
            else:
                instruction = "FillEnumString"
                formtype = ak._v2.forms.IndexedForm

            instructions.append([instruction, index + "-index", "int64", strings])

            return formtype(
                "i64",
                ak._v2.forms.ListOffsetForm(
                    "i64",
                    ak._v2.forms.NumpyForm(
                        "uint8", parameters={"__array__": "char"}, form_key=node
                    ),
                    parameters={"__array__": "string"},
                    form_key=offsets,
                ),
                parameters={"__array__": "categorical"},
                form_key=index,
            )

        else:
            if is_optional:
                mask = f"node{container}"
                container[mask + "-mask"] = None
                instructions.append(["FillByteMaskedArray", mask + "-mask", "int8"])

            offsets = f"node{len(container)}"
            container[offsets + "-offsets"] = None
            node = f"node{len(container)}"
            container[node + "-data"] = None
            instructions.append(
                ["FillString", offsets + "-offsets", "int64", node + "-data", "uint8"]
            )

            out = ak._v2.forms.ListOffsetForm(
                "i64",
                ak._v2.forms.NumpyForm(
                    "uint8",
                    parameters={"__array__": "char"},
                    form_key=node,
                ),
                parameters={"__array__": "string"},
                form_key=offsets,
            )
            if is_optional:
                return ak._v2.forms.ByteMaskedForm(
                    "i8", out, valid_when=True, form_key=mask
                )
            else:
                return out

    elif tpe == "array":
        # https://json-schema.org/understanding-json-schema/reference/array.html

        if "items" not in schema:
            raise TypeError("JSONSchema type is not concrete: array without 'items'")

        if schema.get("minItems") == schema.get("maxItems") != None:  # noqa: E711
            assert ak._v2._util.isint(schema.get("minItems"))

            if is_optional:
                mask = f"node{len(container)}"
                container[mask + "-index"] = None
                instructions.append(
                    ["FillIndexedOptionArray", mask + "-index", "int64"]
                )

            instructions.append(["FixedLengthList", schema.get("minItems")])

            content = build_assembly(schema["items"], container, instructions)

            out = ak._v2.forms.RegularForm(content, size=schema.get("minItems"))
            if is_optional:
                return ak._v2.forms.IndexedOptionForm("i64", out, form_key=mask)
            else:
                return out

        else:
            if is_optional:
                mask = f"node{len(container)}"
                container[mask + "-mask"] = None
                instructions.append(["FillByteMaskedArray", mask + "-mask", "int8"])

            offsets = f"node{len(container)}"
            container[offsets + "-offsets"] = None
            instructions.append(["VarLengthList", offsets + "-offsets", "int64"])

            content = build_assembly(schema["items"], container, instructions)

            out = ak._v2.forms.ListOffsetForm("i64", content, form_key=offsets)
            if is_optional:
                return ak._v2.forms.ByteMaskedForm(
                    "i8", out, valid_when=True, form_key=mask
                )
            else:
                return out

    elif tpe == "object":
        # https://json-schema.org/understanding-json-schema/reference/object.html

        if "properties" not in schema:
            raise TypeError(
                "JSONSchema type is not concrete: object without 'properties'"
            )

        names = []
        subschemas = []
        for name, subschema in schema["properties"].items():
            names.append(name)
            subschemas.append(subschema)

        if is_optional:
            mask = f"node{len(container)}"
            container[mask + "-index"] = None
            instructions.append(["FillIndexedOptionArray", mask + "-index", "int64"])

        instructions.append(["KeyTableHeader", len(names)])
        startkeys = len(instructions)

        for name in names:
            instructions.append(["KeyTableItem", name, None])

        contents = []
        for keyindex, subschema in enumerate(subschemas):
            # set the "jump_to" instruction position in the KeyTable
            instructions[startkeys + keyindex][2] = len(instructions)
            contents.append(build_assembly(subschema, container, instructions))

        out = ak._v2.forms.RecordForm(contents, names)
        if is_optional:
            return ak._v2.forms.IndexedOptionForm("i64", out, form_key=mask)
        else:
            return out

    elif isinstance(tpe, list):
        raise NotImplementedError("arbitrary unions of types are not yet supported")

    else:
        raise TypeError(f"unrecognized JSONSchema: {tpe!r}")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


def from_numpy(
    array, regulararray=False, recordarray=True, highlevel=True, behavior=None
):
    """
    Args:
        array (np.ndarray): The NumPy array to convert into an Awkward Array.
            This array can be a np.ma.MaskedArray.
        regulararray (bool): If True and the array is multidimensional,
            the dimensions are represented by nested #ak.layout.RegularArray
            nodes; if False and the array is multidimensional, the dimensions
            are represented by a multivalued #ak.layout.NumpyArray.shape.
            If the array is one-dimensional, this has no effect.
        recordarray (bool): If True and the array is a NumPy structured array
            (dtype.names is not None), the fields are represented by an
            #ak.layout.RecordArray; if False and the array is a structured
            array, the structure is left in the #ak.layout.NumpyArray `format`,
            which some functions do not recognize.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts a NumPy array into an Awkward Array.

    The resulting layout can only involve the following #ak.layout.Content types:

       * #ak.layout.NumpyArray
       * #ak.layout.ByteMaskedArray or #ak.layout.UnmaskedArray if the
         `array` is an np.ma.MaskedArray.
       * #ak.layout.RegularArray if `regulararray=True`.
       * #ak.layout.RecordArray if `recordarray=True`.

    See also #ak.to_numpy and #ak.from_cupy.
    """

    def recurse(array, mask):
        if regulararray and len(array.shape) > 1:
            return ak._v2.contents.RegularArray(
                recurse(array.reshape((-1,) + array.shape[2:]), mask),
                array.shape[1],
                array.shape[0],
            )

        if len(array.shape) == 0:
            array = array.reshape(1)

        if array.dtype.kind == "S":
            asbytes = array.reshape(-1)
            itemsize = asbytes.dtype.itemsize
            starts = numpy.arange(0, len(asbytes) * itemsize, itemsize, dtype=np.int64)
            stops = starts + numpy.char.str_len(asbytes)
            data = ak._v2.contents.ListArray64(
                ak._v2.index.Index64(starts),
                ak._v2.index.Index64(stops),
                ak._v2.contents.NumpyArray(
                    asbytes.view("u1"), parameters={"__array__": "byte"}, nplike=numpy
                ),
                parameters={"__array__": "bytestring"},
            )
            for i in range(len(array.shape) - 1, 0, -1):
                data = ak._v2.contents.RegularArray(
                    data, array.shape[i], array.shape[i - 1]
                )

        elif array.dtype.kind == "U":
            asbytes = numpy.char.encode(array.reshape(-1), "utf-8", "surrogateescape")
            itemsize = asbytes.dtype.itemsize
            starts = numpy.arange(0, len(asbytes) * itemsize, itemsize, dtype=np.int64)
            stops = starts + numpy.char.str_len(asbytes)
            data = ak._v2.contents.ListArray64(
                ak._v2.index.Index64(starts),
                ak._v2.index.Index64(stops),
                ak._v2.contents.NumpyArray(
                    asbytes.view("u1"), parameters={"__array__": "char"}, nplike=numpy
                ),
                parameters={"__array__": "string"},
            )
            for i in range(len(array.shape) - 1, 0, -1):
                data = ak._v2.contents.RegularArray(
                    data, array.shape[i], array.shape[i - 1]
                )

        else:
            data = ak._v2.contents.NumpyArray(array)

        if mask is None:
            return data

        elif mask is False or (isinstance(mask, np.bool_) and not mask):
            # NumPy's MaskedArray with mask == False is an UnmaskedArray
            if len(array.shape) == 1:
                return ak._v2.contents.UnmaskedArray(data)
            else:

                def attach(x):
                    if isinstance(x, ak._v2.contents.NumpyArray):
                        return ak._v2.contents.UnmaskedArray(x)
                    else:
                        return ak._v2.contents.RegularArray(
                            attach(x.content), x.size, len(x)
                        )

                return attach(data.toRegularArray())

        else:
            # NumPy's MaskedArray is a ByteMaskedArray with valid_when=False
            return ak._v2.contents.ByteMaskedArray(
                ak._v2.index.Index8(mask), data, valid_when=False
            )

    if isinstance(array, numpy.ma.MaskedArray):
        mask = numpy.ma.getmask(array)
        array = numpy.ma.getdata(array)
        if isinstance(mask, np.ndarray) and len(mask.shape) > 1:
            regulararray = True
            mask = mask.reshape(-1)
    else:
        mask = None

    if not recordarray or array.dtype.names is None:
        layout = recurse(array, mask)

    else:
        contents = []
        for name in array.dtype.names:
            contents.append(recurse(array[name], mask))
        layout = ak._v2.contents.RecordArray(contents, array.dtype.names)

    return ak._v2._util.wrap(layout, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak._v2.operations.structure.values_astype(
        ak._v2.highlevel.Array([1.1, 2.2, None, 3.3]), np.float32
    )
    assert (
        str(ak._v2.operations.structure.fill_none(array, np.float32(0)).type)
        == "4 * float32"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_argsort():
    array = ak._v2.Array([[1.1, 2.2], [3.3, 3.1]])
    assert ak._v2.operations.structure.argsort(array).tolist() == [[0, 1], [1, 0]]
    assert (
        str(ak._v2.operations.describe.type(ak._v2.operations.structure.argsort(array)))
        == "2 * var * int64"
    )

    empty_array = ak._v2.Array([[], []])
    assert ak._v2.operations.structure.argsort(empty_array).tolist() == [[], []]
    assert (
        str(
            ak._v2.operations.describe.type(
                ak._v2.operations.structure.argsort(empty_array)
            )
        )
        == "2 * var * int64"
    )

    select_array = array[array > 5]
    assert select_array.tolist() == [[], []]
    assert str(ak._v2.operations.describe.type(select_array)) == "2 * var * float64"

    assert ak._v2.operations.structure.argsort(select_array).tolist() == [[], []]
    assert (
        str(
            ak._v2.operations.describe.type(
                ak._v2.operations.structure.argsort(select_array)
            )
        )
        == "2 * var * int64"
    )

    assert ak._v2.operations.structure.argsort(array[array > 5]).tolist() == [[], []]
    assert (
        str(
            ak._v2.operations.describe.type(
                ak._v2.operations.structure.argsort(array[array > 5])
            )
        )
        == "2 * var * int64"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from awkward._v2.operations.structure.ak_copy import copy  # noqa: F401
from awkward._v2.operations.structure.ak_mask import mask  # noqa: F401
from awkward._v2.operations.structure.ak_num import num  # noqa: F401
from awkward._v2.operations.structure.ak_run_lengths import run_lengths  # noqa: F401
from awkward._v2.operations.structure.ak_zip import zip  # noqa: F401
from awkward._v2.operations.structure.ak_unzip import unzip  # noqa: F401
from awkward._v2.operations.structure.ak_to_regular import to_regular  # noqa: F401
from awkward._v2.operations.structure.ak_from_regular import from_regular  # noqa: F401
from awkward._v2.operations.structure.ak_with_name import with_name  # noqa: F401
from awkward._v2.operations.structure.ak_with_field import with_field  # noqa: F401
from awkward._v2.operations.structure.ak_with_parameter import (  # noqa: F401
    with_parameter,
)
from awkward._v2.operations.structure.ak_without_parameters import (  # noqa: F401
    without_parameters,
)
from awkward._v2.operations.structure.ak_zeros_like import zeros_like  # noqa: F401
from awkward._v2.operations.structure.ak_ones_like import ones_like  # noqa: F401
from awkward._v2.operations.structure.ak_full_like import full_like  # noqa: F401
from awkward._v2.operations.structure.ak_broadcast_arrays import (  # noqa: F401
    broadcast_arrays,
)
from awkward._v2.operations.structure.ak_concatenate import concatenate  # noqa: F401
from awkward._v2.operations.structure.ak_where import where  # noqa: F401
from awkward._v2.operations.structure.ak_flatten import flatten  # noqa: F401
from awkward._v2.operations.structure.ak_unflatten import unflatten  # noqa: F401
from awkward._v2.operations.structure.ak_ravel import ravel  # noqa: F401
from awkward._v2.operations.structure.ak_packed import packed  # noqa: F401
from awkward._v2.operations.structure.ak_local_index import local_index  # noqa: F401
from awkward._v2.operations.structure.ak_sort import sort  # noqa: F401
from awkward._v2.operations.structure.ak_argsort import argsort  # noqa: F401
from awkward._v2.operations.structure.ak_pad_none import pad_none  # noqa: F401
from awkward._v2.operations.structure.ak_fill_none import fill_none  # noqa: F401
from awkward._v2.operations.structure.ak_is_none import is_none  # noqa: F401
from awkward._v2.operations.structure.ak_singletons import singletons  # noqa: F401
from awkward._v2.operations.structure.ak_firsts import firsts  # noqa: F401
from awkward._v2.operations.structure.ak_cartesian import cartesian  # noqa: F401
from awkward._v2.operations.structure.ak_argcartesian import argcartesian  # noqa: F401
from awkward._v2.operations.structure.ak_combinations import combinations  # noqa: F401
from awkward._v2.operations.structure.ak_argcombinations import (  # noqa: F401
    argcombinations,
)
from awkward._v2.operations.structure.ak_nan_to_num import nan_to_num  # noqa: F401
from awkward._v2.operations.structure.ak_isclose import isclose  # noqa: F401
from awkward._v2.operations.structure.ak_values_astype import (  # noqa: F401
    values_astype,
)
from awkward._v2.operations.structure.ak_strings_astype import (  # noqa: F401
    strings_astype,
)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from awkward._v2.operations.reducers.ak_count import count  # noqa: F401
from awkward._v2.operations.reducers.ak_count_nonzero import count_nonzero  # noqa: F401
from awkward._v2.operations.reducers.ak_sum import sum  # noqa: F401
from awkward._v2.operations.reducers.ak_prod import prod  # noqa: F401
from awkward._v2.operations.reducers.ak_any import any  # noqa: F401
from awkward._v2.operations.reducers.ak_all import all  # noqa: F401
from awkward._v2.operations.reducers.ak_min import min  # noqa: F401
from awkward._v2.operations.reducers.ak_max import max  # noqa: F401
from awkward._v2.operations.reducers.ak_ptp import ptp  # noqa: F401
from awkward._v2.operations.reducers.ak_argmin import argmin  # noqa: F401
from awkward._v2.operations.reducers.ak_argmax import argmax  # noqa: F401
from awkward._v2.operations.reducers.ak_moment import moment  # noqa: F401
from awkward._v2.operations.reducers.ak_mean import mean  # noqa: F401
from awkward._v2.operations.reducers.ak_var import var  # noqa: F401
from awkward._v2.operations.reducers.ak_std import std  # noqa: F401
from awkward._v2.operations.reducers.ak_covar import covar  # noqa: F401
from awkward._v2.operations.reducers.ak_corr import corr  # noqa: F401
from awkward._v2.operations.reducers.ak_linear_fit import linear_fit  # noqa: F401
from awkward._v2.operations.reducers.ak_softmax import softmax  # noqa: F401



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy

import awkward as ak
from awkward._v2._util import numpy_at_least
from awkward._v2.contents.numpyarray import NumpyArray

# NumPy 1.13.1 introduced NEP13, without which Awkward ufuncs won't work, which
# would be worse than lacking a feature: it would cause unexpected output.
# NumPy 1.17.0 introduced NEP18, which is optional (use ak.* instead of np.*).
if not numpy_at_least("1.13.1"):
    raise ImportError("NumPy 1.13.1 or later required")


# def convert_to_array(layout, args, kwargs):
#     out = ak._v2.operations.convert.to_numpy(layout, allow_missing=False)
#     if args == () and kwargs == {}:
#         return out
#     else:
#         return numpy.array(out, *args, **kwargs)


# implemented = {}


# def _to_rectilinear(arg):
#     if isinstance(
#         arg,
#         (
#             ak.Array,
#             ak.Record,
#             ak.ArrayBuilder,
#             ak._v2.contents.Content,
#             ak._v2.record.Record,
#             ak.layout.ArrayBuilder,
#             ak.layout.LayoutBuilder32,
#             ak.layout.LayoutBuilder64,
#         ),
#     ):
#         nplike = ak.nplike.of(arg)
#         return nplike.to_rectilinear(arg)
#     else:
#         return arg


# def array_function(func, types, args, kwargs):
#     function = implemented.get(func)
#     if function is None:
#         args = tuple(_to_rectilinear(x) for x in args)
#         kwargs = dict((k, _to_rectilinear(v)) for k, v in kwargs.items())
#         out = func(*args, **kwargs)
#         nplike = ak.nplike.of(out)
#         if isinstance(out, nplike.ndarray) and len(out.shape) != 0:
#             return ak.Array(out)
#         else:
#             return out
#     else:
#         return function(*args, **kwargs)


# def implements(numpy_function):
#     def decorator(function):
#         implemented[getattr(numpy, numpy_function)] = function
#         return function

#     return decorator


def _array_ufunc_custom_cast(inputs, behavior):
    nextinputs = []
    for x in inputs:
        cast_fcn = ak._v2._util.custom_cast(x, behavior)
        if cast_fcn is not None:
            x = cast_fcn(x)
        nextinputs.append(
            ak._v2.operations.convert.to_layout(x, allow_record=True, allow_other=True)
        )
    return nextinputs


def _array_ufunc_adjust(custom, inputs, kwargs, behavior):
    args = [
        ak._v2._util.wrap(x, behavior)
        if isinstance(x, (ak._v2.contents.Content, ak._v2.record.Record))
        else x
        for x in inputs
    ]
    out = custom(*args, **kwargs)
    if not isinstance(out, tuple):
        out = (out,)

    return tuple(
        x.layout
        if isinstance(x, (ak._v2.highlevel.Array, ak._v2.highlevel.Record))
        else x
        for x in out
    )


def _array_ufunc_adjust_apply(apply_ufunc, ufunc, method, inputs, kwargs, behavior):
    nextinputs = [
        ak._v2._util.wrap(x, behavior)
        if isinstance(x, (ak._v2.contents.Content, ak._v2.record.Record))
        else x
        for x in inputs
    ]

    out = apply_ufunc(ufunc, method, nextinputs, kwargs)

    if out is NotImplemented:
        return None
    else:
        if not isinstance(out, tuple):
            out = (out,)
        return tuple(
            x.layout
            if isinstance(x, (ak._v2.highlevel.Array, ak._v2.highlevel.Record))
            else x
            for x in out
        )


def _array_ufunc_signature(ufunc, inputs):
    signature = [ufunc]
    for x in inputs:
        if isinstance(x, ak._v2.contents.Content):
            record, array = x.parameter("__record__"), x.parameter("__array__")
            if record is not None:
                signature.append(record)
            elif array is not None:
                signature.append(array)
            elif isinstance(x, NumpyArray):
                signature.append(x.dtype.type)
            else:
                signature.append(None)
        else:
            signature.append(type(x))

    return signature


def _array_ufunc_deregulate(inputs):
    nextinputs = []
    for x in inputs:
        if isinstance(x, ak._v2.contents.RegularArray):
            y = x.maybe_toNumpyArray()
            if y is not None:
                nextinputs.append(y)
            else:
                nextinputs.append(x)
        else:
            nextinputs.append(x)

    return nextinputs


def array_ufunc(ufunc, method, inputs, kwargs):
    if method != "__call__" or len(inputs) == 0 or "out" in kwargs:
        return NotImplemented

    behavior = ak._v2._util.behavior_of(*inputs)

    inputs = _array_ufunc_custom_cast(inputs, behavior)

    def action(inputs, **ignore):
        signature = _array_ufunc_signature(ufunc, inputs)

        custom = ak._v2._util.overload(behavior, signature)
        if custom is not None:
            return _array_ufunc_adjust(custom, inputs, kwargs, behavior)

        if ufunc is numpy.matmul:
            custom_matmul = action_for_matmul(inputs)
            if custom_matmul is not None:
                return custom_matmul()

        inputs = _array_ufunc_deregulate(inputs)

        if all(
            isinstance(x, NumpyArray) or not isinstance(x, ak._v2.contents.Content)
            for x in inputs
        ):
            nplike = ak.nplike.of(*inputs)

            if nplike.known_data:
                args = []
                for x in inputs:
                    if isinstance(x, NumpyArray):
                        args.append(x.raw(nplike))
                    else:
                        args.append(x)
                result = getattr(ufunc, method)(*args, **kwargs)

            else:
                shape = None
                args = []
                for x in inputs:
                    if isinstance(x, NumpyArray):
                        shape = x.shape
                        args.append(numpy.empty((0,) + x.shape[1:], x.dtype))
                    else:
                        args.append(x)
                assert shape is not None
                tmp = getattr(ufunc, method)(*args, **kwargs)
                result = nplike.empty((shape[0],) + tmp.shape[1:], tmp.dtype)

            return (NumpyArray(result, nplike=nplike),)

        for x in inputs:
            if isinstance(x, ak._v2.contents.Content):
                apply_ufunc = ak._v2._util.custom_ufunc(ufunc, x, behavior)
                if apply_ufunc is not None:
                    out = _array_ufunc_adjust_apply(
                        apply_ufunc, ufunc, method, inputs, kwargs, behavior
                    )
                    if out is not None:
                        return out

        if all(
            x.parameter("__array__") is not None
            or x.parameter("__record__") is not None
            for x in inputs
            if isinstance(x, ak._v2.contents.Content)
        ):
            error_message = []
            for x in inputs:
                if isinstance(x, ak._v2.contents.Content):
                    if x.parameter("__array__") is not None:
                        error_message.append(x.parameter("__array__"))
                    elif x.parameter("__record__") is not None:
                        error_message.append(x.parameter("__record__"))
                    else:
                        error_message.append(type(x).__name__)
                else:
                    error_message.append(type(x).__name__)
            raise TypeError(
                "no {}.{} overloads for custom types: {}".format(
                    type(ufunc).__module__, ufunc.__name__, ", ".join(error_message)
                )
            )

        return None

    if sum(int(isinstance(x, ak._v2.contents.Content)) for x in inputs) == 1:
        where = None
        for i, x in enumerate(inputs):
            if isinstance(x, ak._v2.contents.Content):
                where = i
                break
        assert where is not None

        nextinputs = list(inputs)

        def unary_action(layout, **ignore):
            nextinputs[where] = layout
            result = action(tuple(nextinputs), **ignore)
            if result is None:
                return None
            else:
                assert isinstance(result, tuple) and len(result) == 1
                return result[0]

        out = inputs[where].recursively_apply(
            unary_action, function_name=ufunc.__name__
        )

    else:
        out = ak._v2._broadcasting.broadcast_and_apply(
            inputs, action, behavior, allow_records=False, function_name=ufunc.__name__
        )
        assert isinstance(out, tuple) and len(out) == 1
        out = out[0]

    return ak._v2._util.wrap(out, behavior)


# def matmul_for_numba(lefts, rights, dtype):
#     total_outer = 0
#     total_inner = 0
#     total_content = 0

#     for A, B in zip(lefts, rights):
#         first = -1
#         for Ai in A:
#             if first == -1:
#                 first = len(Ai)
#             elif first != len(Ai):
#                 raise ValueError(
#                     "one of the left matrices in np.matmul is not rectangular"
#                 )
#         if first == -1:
#             first = 0
#         rowsA = len(A)
#         colsA = first

#         first = -1
#         for Bi in B:
#             if first == -1:
#                 first = len(Bi)
#             elif first != len(Bi):
#                 raise ValueError(
#                     "one of the right matrices in np.matmul is not rectangular"
#                 )
#         if first == -1:
#             first = 0
#         rowsB = len(B)
#         colsB = first

#         if colsA != rowsB:
#             raise ValueError(
#                 u"one of the pairs of matrices in np.matmul do not match shape: "
#                 u"(n \u00d7 k) @ (k \u00d7 m)"
#             )

#         total_outer += 1
#         total_inner += rowsA
#         total_content += rowsA * colsB

#     outer = numpy.empty(total_outer + 1, numpy.int64)
#     inner = numpy.empty(total_inner + 1, numpy.int64)
#     content = numpy.zeros(total_content, dtype)

#     outer[0] = 0
#     inner[0] = 0
#     outer_i = 1
#     inner_i = 1
#     content_i = 0
#     for A, B in zip(lefts, rights):
#         rows = len(A)
#         cols = 0
#         if len(B) > 0:
#             cols = len(B[0])
#         mids = 0
#         if len(A) > 0:
#             mids = len(A[0])

#         for i in range(rows):
#             for j in range(cols):
#                 for v in range(mids):
#                     pos = content_i + i * cols + j
#                     content[pos] += A[i][v] * B[v][j]

#         outer[outer_i] = outer[outer_i - 1] + rows
#         outer_i += 1
#         for _ in range(rows):
#             inner[inner_i] = inner[inner_i - 1] + cols
#             inner_i += 1
#         content_i += rows * cols

#     return outer, inner, content


# matmul_for_numba.numbafied = None


def action_for_matmul(inputs):
    raise NotImplementedError


# def action_for_matmul(inputs):
#     inputs = [
#         ak._v2._util.recursively_apply(
#             x, (lambda _: _), pass_depth=False, numpy_to_regular=True
#         )
#         if isinstance(x, (ak._v2.contents.Content, ak._v2.record.Record))
#         else x
#         for x in inputs
#     ]

#     if len(inputs) == 2 and all(
#         isinstance(x, ak._v2._util.listtypes)
#         and isinstance(x.content, ak._v2._util.listtypes)
#         and isinstance(x.content.content, NumpyArray)
#         for x in inputs
#     ):
#         ak._v2._connect.numba.register_and_check()
#         import numba

#         if matmul_for_numba.numbafied is None:
#             matmul_for_numba.numbafied = numba.njit(matmul_for_numba)

#         lefts = ak._v2.highlevel.Array(inputs[0])
#         rights = ak._v2.highlevel.Array(inputs[1])
#         dtype = numpy.asarray(lefts[0:0, 0:0, 0:0] + rights[0:0, 0:0, 0:0]).dtype

#         outer, inner, content = matmul_for_numba.numbafied(lefts, rights, dtype)

#         return lambda: (
#             ak._v2.contents.ListOffsetArray64(
#                 ak._v2.index.Index64(outer),
#                 ak._v2.contents.ListOffsetArray64(
#                     ak._v2.index.Index64(inner),
#                     NumpyArray(content),
#                 ),
#             ),
#         )

#     else:
#         return None


try:
    NDArrayOperatorsMixin = numpy.lib.mixins.NDArrayOperatorsMixin

except AttributeError:
    from numpy.core import umath as um

    def _disables_array_ufunc(obj):
        try:
            return obj.__array_ufunc__ is None
        except AttributeError:
            return False

    def _binary_method(ufunc, name):
        def func(self, other):
            if _disables_array_ufunc(other):
                return NotImplemented
            return ufunc(self, other)

        func.__name__ = f"__{name}__"
        return func

    def _reflected_binary_method(ufunc, name):
        def func(self, other):
            if _disables_array_ufunc(other):
                return NotImplemented
            return ufunc(other, self)

        func.__name__ = f"__r{name}__"
        return func

    def _inplace_binary_method(ufunc, name):
        def func(self, other):
            return ufunc(self, other, out=(self,))

        func.__name__ = f"__i{name}__"
        return func

    def _numeric_methods(ufunc, name):
        return (
            _binary_method(ufunc, name),
            _reflected_binary_method(ufunc, name),
            _inplace_binary_method(ufunc, name),
        )

    def _unary_method(ufunc, name):
        def func(self):
            return ufunc(self)

        func.__name__ = f"__{name}__"
        return func

    class NDArrayOperatorsMixin:
        __lt__ = _binary_method(um.less, "lt")
        __le__ = _binary_method(um.less_equal, "le")
        __eq__ = _binary_method(um.equal, "eq")
        __ne__ = _binary_method(um.not_equal, "ne")
        __gt__ = _binary_method(um.greater, "gt")
        __ge__ = _binary_method(um.greater_equal, "ge")

        __add__, __radd__, __iadd__ = _numeric_methods(um.add, "add")
        __sub__, __rsub__, __isub__ = _numeric_methods(um.subtract, "sub")
        __mul__, __rmul__, __imul__ = _numeric_methods(um.multiply, "mul")
        __matmul__, __rmatmul__, __imatmul__ = _numeric_methods(um.matmul, "matmul")
        __truediv__, __rtruediv__, __itruediv__ = _numeric_methods(
            um.true_divide, "truediv"
        )
        __floordiv__, __rfloordiv__, __ifloordiv__ = _numeric_methods(
            um.floor_divide, "floordiv"
        )
        __mod__, __rmod__, __imod__ = _numeric_methods(um.remainder, "mod")
        if hasattr(um, "divmod"):
            __divmod__ = _binary_method(um.divmod, "divmod")
            __rdivmod__ = _reflected_binary_method(um.divmod, "divmod")
        __pow__, __rpow__, __ipow__ = _numeric_methods(um.power, "pow")
        __lshift__, __rlshift__, __ilshift__ = _numeric_methods(um.left_shift, "lshift")
        __rshift__, __rrshift__, __irshift__ = _numeric_methods(
            um.right_shift, "rshift"
        )
        __and__, __rand__, __iand__ = _numeric_methods(um.bitwise_and, "and")
        __xor__, __rxor__, __ixor__ = _numeric_methods(um.bitwise_xor, "xor")
        __or__, __ror__, __ior__ = _numeric_methods(um.bitwise_or, "or")

        __neg__ = _unary_method(um.negative, "neg")
        if hasattr(um, "positive"):
            __pos__ = _unary_method(um.positive, "pos")
        __abs__ = _unary_method(um.absolute, "abs")
        __invert__ = _unary_method(um.invert, "invert")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak


def backend(*arrays):
    """
    Returns the names of the backend used by `arrays`. May be

       * `"cpu"` for `libawkward-cpu-kernels.so`;
       * `"cuda"` for `libawkward-cuda-kernels.so`;
       * `"mixed"` if any of the arrays have different labels within their
         structure or any arrays have different labels from each other;
       * None if the objects are not Awkward, NumPy, or CuPy arrays (e.g.
         Python numbers, booleans, strings).

    Mixed arrays can't be used in any operations, and two arrays on different
    devices can't be used in the same operation.

    To use `"cuda"`, the package
    [awkward-cuda-kernels](https://pypi.org/project/awkward-cuda-kernels)
    be installed, either by

        pip install awkward-cuda-kernels

    or as an optional dependency with

        pip install awkward[cuda] --upgrade

    It is only available for Linux as a binary wheel, and only supports Nvidia
    GPUs (it is written in CUDA).

    See #ak.to_backend.
    """
    backends = set()
    for array in arrays:
        layout = ak._v2.operations.convert.to_layout(
            array,
            allow_record=True,
            allow_other=True,
        )
        if isinstance(layout, (ak._v2.contents.Content, ak._v2.index.Index)):
            if isinstance(layout.nplike, ak.nplike.Numpy):
                backends.add("cpu")
            elif isinstance(layout.nplike, ak.nplike.Cupy):
                backends.add("cuda")
        elif isinstance(layout, ak.nplike.numpy.ndarray):
            backends.add("cpu")
        elif type(layout).__module__.startswith("cupy."):
            backends.add("cuda")

    if backends == set():
        return None
    elif backends == {"cpu"}:
        return "cpu"
    elif backends == {"cuda"}:
        return "cuda"
    else:
        return "mixed"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def without_parameters(array, highlevel=True, behavior=None):

    """
    Args:
        array: Data convertible into an Awkward Array.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    This function returns a new array without any parameters in its
    #ak.Array.layout, on nodes of any level of depth.

    Note that a "new array" is a lightweight shallow copy, not a duplication
    of large data buffers.
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=True, allow_other=False
    )

    out = layout.recursively_apply(lambda layout, **kwargs: None, keep_parameters=False)

    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = np.array([1, 2, 9, 0])
    nplike = ak.nplike.of(array)

    ak_array = ak.from_numpy(array)
    assert ak.to_list(nplike.to_rectilinear(array)) == ak.to_list(ak_array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def test():
    first = ak._v2.operations.convert.from_numpy(np.array([1, 2, 3]))
    deltas = ak._v2.highlevel.Array([[1, 2], [1, 2], [1, 2, 3]])
    assert np.hstack(
        (
            first[:, np.newaxis],
            ak._v2.operations.structure.fill_none(
                ak.operations.structure.pad_none(deltas, 3, axis=-1), 999
            ),
        )
    ).tolist() == [[1, 1, 2, 999], [2, 1, 2, 999], [3, 1, 2, 3]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

# from __future__ import absolute_import

# import sys

# import awkward as ak


# def mixin_class(registry, name=None):
#     """
#     Args:
#         registry (dict): The destination behavior mapping registry. Typically,
#             this would be the global registry #ak.behavior, but one may wish
#             to register methods in an alternative way.
#         name (str): The name to assign to the behaviour class.

#     This decorator can be used to register a behavior mixin class.

#     Any inherited behaviors will automatically be made available to the decorated
#     class.

#     See the "Mixin decorators" section of #ak.behavior for further details.
#     """

#     def register(cls):
#         cls_name = cls.__name__
#         if name is None:
#             behavior_name = cls_name
#         else:
#             behavior_name = name

#         record = type(
#             cls_name + "Record",
#             (cls, ak._v2.highlevel.Record),
#             {"__module__": cls.__module__},
#         )
#         setattr(sys.modules[cls.__module__], cls_name + "Record", record)
#         registry[behavior_name] = record
#         array = type(
#             cls_name + "Array",
#             (cls, ak._v2.highlevel.Array),
#             {"__module__": cls.__module__},
#         )
#         setattr(sys.modules[cls.__module__], cls_name + "Array", array)
#         registry["*", behavior_name] = array
#         for basecls in cls.mro():
#             for method in basecls.__dict__.values():
#                 if hasattr(method, "_awkward_mixin"):
#                     ufunc, rhs, transpose = method._awkward_mixin
#                     if rhs is None:
#                         registry.setdefault((ufunc, behavior_name), method)
#                         continue
#                     for rhs_name in list(rhs) + [behavior_name]:
#                         registry.setdefault((ufunc, behavior_name, rhs_name), method)
#                         if transpose is not None and rhs_name != behavior_name:
#                             registry.setdefault(
#                                 (ufunc, rhs_name, behavior_name), transpose
#                             )
#                     if basecls.__name__ in rhs:
#                         rhs.add(behavior_name)
#         return cls

#     return register


# def mixin_class_method(ufunc, rhs=None, transpose=True):
#     """
#     Args:
#         ufunc (numpy.ufunc): A universal function (or NEP18 callable) that is
#             hooked in Awkward Array, i.e. it can be the first argument of a behavior.
#         rhs (Set[type] or None): Set of right-hand side argument types, optional
#             if wrapping a unary function. The left-hand side is expected to
#             always be `self` of the parent class.
#         transpose (bool): If true, automatically create a transpose signature
#             (only makes sense for binary ufuncs).

#     This decorator can be used to register a mixin class method.

#     Using this decorator ensures that derived classes that are declared with the
#     #ak.mixin_class decorator will also have the behaviors that this class has.
#     """

#     def register(method):
#         if not isinstance(rhs, (set, type(None))):
#             raise ValueError(
#                 "expected a set of right-hand-side argument types"
#
#             )
#         if transpose and rhs is not None:

#             def transposed(left, right):
#                 return method(right, left)

#             # make a copy of rhs, we will edit it later
#             method._awkward_mixin = (ufunc, set(rhs), transposed)
#         else:
#             method._awkward_mixin = (ufunc, rhs, None)
#         return method

#     return register


# __all__ = [
#     x for x in list(globals()) if not x.startswith("_") and x not in ("sys", "ak")
# ]


# def __dir__():
#     return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def from_cupy(array, regulararray=False, highlevel=True, behavior=None):
    """
    Args:
        array (cp.ndarray): The CuPy array to convert into an Awkward Array.
        regulararray (bool): If True and the array is multidimensional,
            the dimensions are represented by nested #ak.layout.RegularArray
            nodes; if False and the array is multidimensional, the dimensions
            are represented by a multivalued #ak.layout.NumpyArray.shape.
            If the array is one-dimensional, this has no effect.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts a CuPy array into an Awkward Array.

    The resulting layout may involve the following #ak.layout.Content types
    (only):

       * #ak.layout.NumpyArray
       * #ak.layout.RegularArray if `regulararray=True`.

    See also #ak.to_cupy, #ak.from_numpy and #ak.from_jax.
    """

    def recurse(array):
        if regulararray and len(array.shape) > 1:
            return ak._v2.contents.RegularArray(
                recurse(array.reshape((-1,) + array.shape[2:])),
                array.shape[1],
                array.shape[0],
            )

        if len(array.shape) == 0:
            data = ak._v2.contents.NumpyArray(array.reshape(1))
        else:
            data = ak._v2.contents.NumpyArray(array)

        return data

    layout = recurse(array)

    return ak._v2._util.wrap(layout, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.numpyform import NumpyForm
from awkward._v2.forms.form import _parameters_equal
from awkward._v2.types.numpytype import primitive_to_dtype

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


class NumpyArray(Content):
    is_NumpyType = True

    def __init__(self, data, identifier=None, parameters=None, nplike=None):
        if nplike is None:
            nplike = ak.nplike.of(data)
        if isinstance(data, ak._v2.index.Index):
            data = data.data
        self._data = nplike.asarray(data)

        ak._v2.types.numpytype.dtype_to_primitive(self._data.dtype)
        if len(self._data.shape) == 0:
            raise TypeError(
                "{} 'data' must be an array, not {}".format(
                    type(self).__name__, repr(data)
                )
            )

        self._init(identifier, parameters, nplike)

    @property
    def data(self):
        return self._data

    @property
    def shape(self):
        return self._data.shape

    @property
    def inner_shape(self):
        return self._data.shape[1:]

    @property
    def strides(self):
        return self._data.strides

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def ptr(self):
        return self._data.ctypes.data

    def raw(self, nplike):
        return self.nplike.raw(self.data, nplike)

    Form = NumpyForm

    def _form_with_key(self, getkey):
        return self.Form(
            ak._v2.types.numpytype.dtype_to_primitive(self._data.dtype),
            self._data.shape[1:],
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=getkey(self),
        )

    def _to_buffers(self, form, getkey, container, nplike):
        assert isinstance(form, self.Form)
        key = getkey(self, form, "data")
        container[key] = ak._v2._util.little_endian(self.raw(nplike))

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return NumpyArray(
            self.raw(tt),
            self._typetracer_identifier(),
            self._parameters,
            tt,
        )

    @property
    def length(self):
        return self._data.shape[0]

    def _forget_length(self):
        return NumpyArray(
            self._data.forget_length(),
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<NumpyArray dtype="]
        out.append(repr(str(self.dtype)))
        if len(self._data.shape) == 1:
            out.append(" len=" + repr(str(self._data.shape[0])))
        else:
            out.append(
                " shape='({})'".format(", ".join(str(x) for x in self._data.shape))
            )

        extra = self._repr_extra(indent + "    ")
        arraystr_lines = self._nplike.array_str(self._data, max_line_width=30).split(
            "\n"
        )
        if len(extra) != 0 or len(arraystr_lines) > 1:
            arraystr_lines = self._nplike.array_str(
                self._data, max_line_width=max(80 - len(indent) - 4, 40)
            ).split("\n")
            if len(arraystr_lines) > 5:
                arraystr_lines = arraystr_lines[:2] + [" ..."] + arraystr_lines[-2:]
            out.append(">")
            out.extend(extra)
            out.append("\n" + indent + "    ")
            out.append(("\n" + indent + "    ").join(arraystr_lines))
            out.append("\n" + indent + "</NumpyArray>")
        else:
            out.append(">")
            out.append(arraystr_lines[0])
            out.append("</NumpyArray>")

        out.append(post)
        return "".join(out)

    def merge_parameters(self, parameters):
        return NumpyArray(
            self._data,
            self._identifier,
            ak._v2._util.merge_parameters(self._parameters, parameters),
            self._nplike,
        )

    def toRegularArray(self):
        shape = self._data.shape
        zeroslen = [1]
        for x in shape:
            zeroslen.append(zeroslen[-1] * x)

        out = NumpyArray(self._data.reshape(-1), None, None, self._nplike)
        for i in range(len(shape) - 1, 0, -1):
            out = ak._v2.contents.RegularArray(
                out, shape[i], zeroslen[i], None, None, self._nplike
            )
        out._identifier = self._identifier
        out._parameters = self._parameters
        return out

    def _nonfinite_to_union(self, nan_string, infinity_string, minus_infinity_string):
        shape = self._data.shape
        zeroslen = [1]
        for x in shape:
            zeroslen.append(zeroslen[-1] * x)

        out = NumpyArray(self._data.reshape(-1), None, None, self._nplike)

        is_nonfinite = ~self._nplike.isfinite(self._data)  # true for inf, -inf, nan
        is_posinf = is_nonfinite & (self._data > 0)  # true for inf only
        is_neginf = is_nonfinite & (self._data < 0)  # true for -inf only
        is_nan = self._nplike.isnan(self._data)  # true for nan only
        tags = self._nplike.zeros(out.length, np.int8)
        tags[is_nonfinite] = 1
        index = self._nplike.arange(out.length, dtype=np.int64)
        index[is_posinf] = 0
        index[is_neginf] = 1
        index[is_nan] = 2

        out = ak._v2.contents.unionarray.UnionArray(
            tags=ak._v2.index.Index8(tags),
            index=ak._v2.index.Index64(index),
            contents=[
                out,
                ak._v2.operations.convert.from_iter(
                    [
                        infinity_string if infinity_string is not None else "Infinity",
                        minus_infinity_string
                        if minus_infinity_string is not None
                        else "-Infinity",
                        nan_string if nan_string is not None else "NaN",
                    ],
                    highlevel=False,
                ),
            ],
        )
        for i in range(len(shape) - 1, 0, -1):
            out = ak._v2.contents.RegularArray(
                out, shape[i], zeroslen[i], None, None, self._nplike
            )
        out._identifier = self._identifier
        out._parameters = self._parameters

        return out

    def maybe_to_array(self, nplike):
        return nplike.asarray(self._data)

    def __array__(self, **kwargs):
        return numpy.asarray(self._data, **kwargs)

    def __iter__(self):
        return iter(self._data)

    def _getitem_nothing(self):
        tmp = self._data[0:0]
        return NumpyArray(
            tmp.reshape((0,) + tmp.shape[2:]),
            self._range_identifier(0, 0),
            None,
            self._nplike,
        )

    def _getitem_at(self, where):
        if not self._nplike.known_data and len(self._data.shape) == 1:
            return ak._v2._typetracer.UnknownScalar(self._data.dtype)

        try:
            out = self._data[where]
        except IndexError as err:
            raise NestedIndexError(self, where, str(err))

        if hasattr(out, "shape") and len(out.shape) != 0:
            return NumpyArray(out, None, None, self._nplike)
        else:
            return out

    def _getitem_range(self, where):
        if not self._nplike.known_shape:
            return self

        start, stop, step = where.indices(self.length)
        assert step == 1

        try:
            out = self._data[where]
        except IndexError as err:
            raise NestedIndexError(self, where, str(err))

        return NumpyArray(
            out,
            self._range_identifier(start, stop),
            self._parameters,
            self._nplike,
        )

    def _getitem_field(self, where, only_fields=()):
        raise NestedIndexError(self, where, "not an array of records")

    def _getitem_fields(self, where, only_fields=()):
        if len(where) == 0:
            return self._getitem_range(slice(0, 0))
        raise NestedIndexError(self, where, "not an array of records")

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)
        try:
            nextdata = self._data[carry.data]
        except IndexError as err:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, str(err))
            else:
                raise exception(str(err))
        return NumpyArray(
            nextdata,
            self._carry_identifier(carry, exception),
            self._parameters,
            self._nplike,
        )

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        if self._data.ndim == 1:
            raise NestedIndexError(
                self,
                ak._v2.contents.ListArray(
                    slicestarts, slicestops, slicecontent, None, None, self._nplike
                ),
                "too many jagged slice dimensions for array",
            )
        else:
            next = self.toRegularArray()
            return next._getitem_next_jagged(
                slicestarts, slicestops, slicecontent, tail
            )

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, int):
            where = (slice(None), head) + tail

            try:
                out = self._data[where]
            except IndexError as err:
                raise NestedIndexError(self, (head,) + tail, str(err))

            if hasattr(out, "shape") and len(out.shape) != 0:
                return NumpyArray(out, None, None, self._nplike)
            else:
                return out

        elif isinstance(head, slice) or head is np.newaxis or head is Ellipsis:
            where = (slice(None), head) + tail
            try:
                out = self._data[where]
            except IndexError as err:
                raise NestedIndexError(self, (head,) + tail, str(err))
            out2 = NumpyArray(out, None, self._parameters, self._nplike)
            return out2

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif isinstance(head, ak._v2.index.Index64):
            if advanced is None:
                where = (slice(None), head.data) + tail
            else:
                where = (self._nplike.asarray(advanced.data), head.data) + tail

            try:
                out = self._data[where]
            except IndexError as err:
                raise NestedIndexError(self, (head,) + tail, str(err))

            return NumpyArray(out, None, self._parameters, self._nplike)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            where = (slice(None), head) + tail
            try:
                out = self._data[where]
            except IndexError as err:
                raise NestedIndexError(self, (head,) + tail, str(err))
            out2 = NumpyArray(out, None, self._parameters, self._nplike)
            return out2

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            next = self.toRegularArray()
            return next._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def num(self, axis, depth=0):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            out = ak._v2.index.Index64.empty(1, self._nplike)
            out[0] = self.length
            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)[
                0
            ]
        shape = []
        reps = 1
        size = self.length
        i = 0
        while i < self._data.ndim - 1 and depth < posaxis:
            shape.append(self.shape[i])
            reps *= self.shape[i]
            size = self.shape[i + 1]
            i += 1
            depth += 1
        if posaxis > depth:
            raise np.AxisError(f"axis={axis} exceeds the depth of this array ({depth})")

        tonum = ak._v2.index.Index64.empty(reps, self._nplike)
        assert tonum.nplike is self._nplike
        self._handle_error(
            self._nplike["awkward_RegularArray_num", tonum.dtype.type](
                tonum.data, size, reps
            )
        )
        return ak._v2.contents.numpyarray.NumpyArray(
            tonum.data.reshape(shape), None, self.parameters, self._nplike
        )

    def _offsets_and_flattened(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            raise np.AxisError(self, "axis=0 not allowed for flatten")

        elif len(self.shape) != 1:
            return self.toRegularArray()._offsets_and_flattened(posaxis, depth)

        else:
            raise np.AxisError(self, "axis out of range for flatten")

    def mergeable(self, other, mergebool):
        if not _parameters_equal(self._parameters, other._parameters):
            return False

        if isinstance(
            other,
            (
                ak._v2.contents.emptyarray.EmptyArray,
                ak._v2.contents.unionarray.UnionArray,
            ),
        ):
            return True

        elif isinstance(
            other,
            (
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return self.mergeable(other._content, mergebool)

        if isinstance(other, ak._v2.contents.numpyarray.NumpyArray):
            if self._data.ndim != other._data.ndim:
                return False

            matching_dtype = self._data.dtype == other._data.dtype

            if (
                not mergebool
                and not matching_dtype
                and (
                    self._data.dtype.type is np.bool_
                    or other._data.dtype.type is np.bool_
                )
            ):
                return False

            if not matching_dtype and np.datetime64 in (
                self._data.dtype,
                other._data.dtype,
            ):
                return False

            if not matching_dtype and np.timedelta64 in (
                self._data.dtype,
                other._data.dtype,
            ):
                return False

            if (
                len(self._data.shape) > 1
                and self._data.shape[1:] != other._data.shape[1:]
            ):
                return False

            return True

        else:
            return False

    def mergemany(self, others):
        if len(others) == 0:
            return self
        head, tail = self._merging_strategy(others)

        contiguous_arrays = []

        for array in head:
            parameters = ak._v2._util.merge_parameters(
                self._parameters, array._parameters, True
            )
            if isinstance(array, ak._v2.contents.emptyarray.EmptyArray):
                pass
            elif isinstance(array, ak._v2.contents.numpyarray.NumpyArray):
                contiguous_arrays.append(array.data)
            else:
                raise AssertionError(
                    "cannot merge "
                    + type(self).__name__
                    + " with "
                    + type(array).__name__
                )

        contiguous_arrays = self._nplike.concatenate(contiguous_arrays)

        next = NumpyArray(contiguous_arrays, self._identifier, parameters, self._nplike)

        if len(tail) == 0:
            return next

        reversed = tail[0]._reverse_merge(next)
        if len(tail) == 1:
            return reversed
        else:
            return reversed.mergemany(tail[1:])

    def fillna(self, value):
        return self

    def _localindex(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        elif len(self.shape) <= 1:
            raise np.AxisError(self, "'axis' out of range for localindex")
        else:
            return self.toRegularArray()._localindex(posaxis, depth)

    def contiguous(self):
        if self.is_contiguous:
            return self
        else:
            return ak._v2.contents.NumpyArray(
                self._nplike.ascontiguousarray(self._data),
                self._identifier,
                self._parameters,
                self._nplike,
            )

    @property
    def is_contiguous(self):
        if isinstance(self._nplike, ak._v2._typetracer.TypeTracer):
            return True

        # Alternatively, self._data.flags["C_CONTIGUOUS"], but the following assumes
        # less of the nplike.

        x = self._data.dtype.itemsize

        for i in range(len(self._data.shape), 0, -1):
            if x != self._data.strides[i - 1]:
                return False
            else:
                x = x * self._data.shape[i - 1]

        return True

    def _subranges_equal(self, starts, stops, length, sorted=True):
        is_equal = ak._v2.index.Index64.zeros(1, self._nplike)

        tmp = self._nplike.empty(length, self.dtype)
        self._handle_error(
            self._nplike[  # noqa: E231
                "awkward_NumpyArray_fill",
                self.dtype.type,
                self._data.dtype.type,
            ](
                tmp,
                0,
                self._data,
                length,
            )
        )

        if not sorted:
            tmp_beg_ptr = ak._v2.index.Index64.empty(ak._util.kMaxLevels, self._nplike)
            tmp_end_ptr = ak._v2.index.Index64.empty(ak._util.kMaxLevels, self._nplike)

            assert (
                tmp_beg_ptr.nplike is self._nplike
                and tmp_end_ptr.nplike is self._nplike
                and starts.nplike is self._nplike
                and stops.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_quick_sort",
                    self.dtype.type,
                    tmp_beg_ptr.dtype.type,
                    tmp_end_ptr.dtype.type,
                    starts.dtype.type,
                    stops.dtype.type,
                ](
                    tmp,
                    tmp_beg_ptr.data,
                    tmp_end_ptr.data,
                    starts.data,
                    stops.data,
                    True,
                    starts.length,
                    ak._util.kMaxLevels,
                )
            )
        assert starts.nplike is self._nplike and stops.nplike is self._nplike
        self._handle_error(
            self._nplike[
                "awkward_NumpyArray_subrange_equal",
                self.dtype.type,
                starts.dtype.type,
                stops.dtype.type,
                np.bool_,
            ](
                tmp,
                starts.data,
                stops.data,
                starts.length,
                is_equal.data,
            )
        )

        return True if is_equal[0] == 1 else False

    def _as_unique_strings(self, offsets):
        outoffsets = ak._v2.index.Index64.empty(offsets.length, self._nplike)
        out = self._nplike.empty(self.shape[0], self.dtype)

        assert offsets.nplike is self._nplike and outoffsets.nplike is self._nplike
        self._handle_error(
            self._nplike[
                "awkward_NumpyArray_sort_asstrings_uint8",
                self.dtype.type,
                self._data.dtype.type,
                offsets._data.dtype.type,
                outoffsets.dtype.type,
            ](
                out,
                self._data,
                offsets.data,
                offsets.length,
                outoffsets.data,
                True,
                False,
            )
        )

        outlength = ak._v2.index.Index64.empty(1, self._nplike)
        nextoffsets = ak._v2.index.Index64.empty(offsets.length, self._nplike)
        assert (
            outoffsets.nplike is self._nplike
            and nextoffsets.nplike is self._nplike
            and outlength.nplike is self._nplike
        )
        self._handle_error(
            self._nplike[
                "awkward_NumpyArray_unique_strings",
                self.dtype.type,
                outoffsets.dtype.type,
                nextoffsets.dtype.type,
                outlength.dtype.type,
            ](
                out,
                outoffsets.data,
                offsets.length,
                nextoffsets.data,
                outlength.data,
            )
        )
        out2 = NumpyArray(out, None, self._parameters, self._nplike)

        return out2, nextoffsets[: outlength[0]]

    def numbers_to_type(self, name):
        if (
            self.parameter("__array__") == "string"
            or self.parameter("__array__") == "bytestring"
            or self.parameter("__array__") == "char"
            or self.parameter("__array__") == "byte"
        ):
            return self
        else:
            dtype = primitive_to_dtype(name)
            return NumpyArray(
                self._nplike.asarray(self._data, dtype=dtype),
                self._identifier,
                self._parameters,
                self._nplike,
            )

    def _is_unique(self, negaxis, starts, parents, outlength):
        if self.length == 0:
            return True

        elif len(self.shape) != 1 or not self.is_contiguous:
            contiguous_self = self if self.is_contiguous else self.contiguous()
            return contiguous_self.toRegularArray()._is_unique(
                negaxis,
                starts,
                parents,
                outlength,
            )
        else:
            out = self._unique(negaxis, starts, parents, outlength)
            if isinstance(out, ak._v2.contents.ListOffsetArray):
                return out.content.length == self.length

            return out.length == self.length

    def _unique(self, negaxis, starts, parents, outlength):
        if self.shape[0] == 0:
            return self

        if len(self.shape) == 0:
            return self

        if negaxis is None:
            contiguous_self = self if self.is_contiguous else self.contiguous()
            # Python 3.8 could use math.prod
            flattened_shape = 1
            for s in contiguous_self.shape:
                flattened_shape = flattened_shape * s

            offsets = ak._v2.index.Index64.zeros(2, self._nplike)
            offsets[1] = flattened_shape
            dtype = (
                np.dtype(np.int64)
                if self._data.dtype.kind.upper() == "M"
                else self._data.dtype
            )
            out = self._nplike.empty(offsets[1], dtype)
            assert offsets.nplike is self._nplike
            self._handle_error(
                self._nplike[  # noqa: E231
                    "awkward_sort",
                    dtype.type,
                    dtype.type,
                    offsets.dtype.type,
                ](
                    out,
                    contiguous_self._data,
                    offsets[1],
                    offsets.data,
                    2,
                    offsets[1],
                    True,
                    False,
                )
            )

            nextlength = ak._v2.index.Index64.empty(1, self._nplike)
            assert nextlength.nplike is self._nplike
            self._handle_error(
                self._nplike[  # noqa: E231
                    "awkward_unique",
                    out.dtype.type,
                    nextlength.dtype.type,
                ](
                    out,
                    out.shape[0],
                    nextlength.data,
                )
            )

            return ak._v2.contents.NumpyArray(
                self._nplike.asarray(out[: nextlength[0]], self.dtype),
                None,
                None,
                self._nplike,
            )

        # axis is not None
        if len(self.shape) != 1 or not self.is_contiguous:
            contiguous_self = self if self.is_contiguous else self.contiguous()
            return contiguous_self.toRegularArray()._unique(
                negaxis,
                starts,
                parents,
                outlength,
            )
        else:

            parents_length = parents.length
            offsets_length = ak._v2.index.Index64.empty(1, self._nplike)
            assert (
                offsets_length.nplike is self._nplike and parents.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_sorting_ranges_length",
                    offsets_length.dtype.type,
                    parents.dtype.type,
                ](
                    offsets_length.data,
                    parents.data,
                    parents_length,
                )
            )

            offsets = ak._v2.index.Index64.empty(offsets_length[0], self._nplike)
            assert offsets.nplike is self._nplike and parents.nplike is self._nplike
            self._handle_error(
                self._nplike[  # noqa: E231
                    "awkward_sorting_ranges",
                    offsets.dtype.type,
                    parents.dtype.type,
                ](
                    offsets.data,
                    offsets_length[0],
                    parents.data,
                    parents_length,
                )
            )

            out = self._nplike.empty(self.length, self.dtype)
            assert offsets.nplike is self._nplike
            self._handle_error(
                self._nplike[
                    "awkward_sort",
                    out.dtype.type,
                    self._data.dtype.type,
                    offsets.dtype.type,
                ](
                    out,
                    self._data,
                    self.shape[0],
                    offsets.data,
                    offsets_length[0],
                    parents_length,
                    True,
                    False,
                )
            )

            nextoffsets = ak._v2.index.Index64.empty(offsets.length, self._nplike)
            assert offsets.nplike is self._nplike and nextoffsets.nplike is self._nplike
            self._handle_error(
                self._nplike[
                    "awkward_unique_ranges",
                    out.dtype.type,
                    offsets.dtype.type,
                    nextoffsets.dtype.type,
                ](
                    out,
                    out.shape[0],
                    offsets.data,
                    offsets.length,
                    nextoffsets.data,
                )
            )

            outoffsets = ak._v2.index.Index64.empty(starts.length + 1, self._nplike)

            assert (
                outoffsets.nplike is self._nplike
                and nextoffsets.nplike is self._nplike
                and starts.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_unique_offsets",
                    outoffsets.dtype.type,
                    nextoffsets.dtype.type,
                    starts.dtype.type,
                ](
                    outoffsets.data,
                    nextoffsets.length,
                    nextoffsets.data,
                    starts.data,
                    starts.length,
                )
            )

            return ak._v2.contents.ListOffsetArray(
                outoffsets,
                ak._v2.contents.NumpyArray(out),
                None,
                self._parameters,
                self._nplike,
            )

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self.shape[0] == 0:
            return ak._v2.contents.NumpyArray(
                self._nplike.empty(0, np.int64), None, None, self._nplike
            )

        if len(self.shape) == 0:
            raise TypeError(f"{type(self).__name__} attempting to argsort a scalar ")
        elif len(self.shape) != 1 or not self.is_contiguous:
            contiguous_self = self if self.is_contiguous else self.contiguous()
            return contiguous_self.toRegularArray()._argsort_next(
                negaxis,
                starts,
                shifts,
                parents,
                outlength,
                ascending,
                stable,
                kind,
                order,
            )

        else:
            parents_length = parents.length
            offsets_length = ak._v2.index.Index64.empty(1, self._nplike)
            assert (
                offsets_length.nplike is self._nplike and parents.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_sorting_ranges_length",
                    offsets_length.dtype.type,
                    parents.dtype.type,
                ](
                    offsets_length.data,
                    parents.data,
                    parents_length,
                )
            )
            offsets_length = offsets_length[0]

            offsets = ak._v2.index.Index64.empty(offsets_length, self._nplike)
            assert offsets.nplike is self._nplike and parents.nplike is self._nplike
            self._handle_error(
                self._nplike[
                    "awkward_sorting_ranges",
                    offsets.dtype.type,
                    parents.dtype.type,
                ](
                    offsets.data,
                    offsets_length,
                    parents.data,
                    parents_length,
                )
            )

            dtype = (
                np.dtype(np.int64)
                if self._data.dtype.kind.upper() == "M"
                else self._data.dtype
            )
            nextcarry = ak._v2.index.Index64.empty(self.__len__(), self._nplike)
            assert nextcarry.nplike is self._nplike and offsets.nplike is self._nplike
            self._handle_error(
                self._nplike[
                    "awkward_argsort",
                    nextcarry.dtype.type,
                    dtype.type,
                    offsets.dtype.type,
                ](
                    nextcarry.data,
                    self._data,
                    self.__len__(),
                    offsets.data,
                    offsets_length,
                    ascending,
                    stable,
                )
            )

            if shifts is not None:
                assert (
                    nextcarry.nplike is self._nplike
                    and shifts.nplike is self._nplike
                    and offsets.nplike is self._nplike
                    and parents.nplike is self._nplike
                    and starts.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_NumpyArray_rearrange_shifted",
                        nextcarry.dtype.type,
                        shifts.dtype.type,
                        offsets.dtype.type,
                        parents.dtype.type,
                        starts.dtype.type,
                    ](
                        nextcarry.data,
                        shifts.data,
                        shifts.length,
                        offsets.data,
                        offsets_length,
                        parents.data,
                        parents_length,
                        starts.data,
                        starts.length,
                    )
                )
            out = NumpyArray(nextcarry, None, None, self._nplike)
            return out

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        if len(self.shape) == 0:
            raise TypeError(f"{type(self).__name__} attempting to sort a scalar ")

        elif len(self.shape) != 1 or not self.is_contiguous:
            contiguous_self = self if self.is_contiguous else self.contiguous()
            return contiguous_self.toRegularArray()._sort_next(
                negaxis,
                starts,
                parents,
                outlength,
                ascending,
                stable,
                kind,
                order,
            )

        else:
            parents_length = parents.length
            offsets_length = ak._v2.index.Index64.empty(1, self._nplike)
            assert (
                offsets_length.nplike is self._nplike and parents.nplike is self._nplike
            )
            self._handle_error(
                self._nplike[
                    "awkward_sorting_ranges_length",
                    offsets_length.dtype.type,
                    parents.dtype.type,
                ](
                    offsets_length.data,
                    parents.data,
                    parents_length,
                )
            )

            offsets = ak._v2.index.Index64.empty(offsets_length[0], self._nplike)

            assert offsets.nplike is self._nplike and parents.nplike is self._nplike
            self._handle_error(
                self._nplike[
                    "awkward_sorting_ranges",
                    offsets.dtype.type,
                    parents.dtype.type,
                ](
                    offsets.data,
                    offsets_length[0],
                    parents.data,
                    parents_length,
                )
            )

            dtype = (
                np.dtype(np.int64)
                if self._data.dtype.kind.upper() == "M"
                else self._data.dtype
            )
            out = self._nplike.empty(self.length, dtype)
            assert offsets.nplike is self._nplike
            self._handle_error(
                self._nplike[  # noqa: E231
                    "awkward_sort",
                    dtype.type,
                    dtype.type,
                    offsets.dtype.type,
                ](
                    out,
                    self._data,
                    self.shape[0],
                    offsets.data,
                    offsets_length[0],
                    parents_length,
                    ascending,
                    stable,
                )
            )
            return ak._v2.contents.NumpyArray(
                self._nplike.asarray(out, self.dtype), None, None, self._nplike
            )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        elif len(self.shape) <= 1:
            raise np.AxisError("'axis' out of range for combinations")
        else:
            return self.toRegularArray()._combinations(
                n, replacement, recordlookup, parameters, posaxis, depth
            )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        if len(self._data.shape) != 1 or not self.is_contiguous:
            return self.toRegularArray()._reduce_next(
                reducer,
                negaxis,
                starts,
                shifts,
                parents,
                outlength,
                mask,
                keepdims,
            )

        out = reducer.apply(self, parents, outlength)

        if reducer.needs_position:
            if shifts is None:
                assert (
                    out.nplike is self._nplike
                    and parents.nplike is self._nplike
                    and starts.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_NumpyArray_reduce_adjust_starts_64",
                        out.data.dtype.type,
                        parents.dtype.type,
                        starts.dtype.type,
                    ](
                        out.data,
                        outlength,
                        parents.data,
                        starts.data,
                    )
                )
            else:
                assert (
                    out.nplike is self._nplike
                    and parents.nplike is self._nplike
                    and starts.nplike is self._nplike
                    and shifts.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_NumpyArray_reduce_adjust_starts_shifts_64",
                        out.data.dtype.type,
                        parents.dtype.type,
                        starts.dtype.type,
                        shifts.dtype.type,
                    ](
                        out.data,
                        outlength,
                        parents.data,
                        starts.data,
                        shifts.data,
                    )
                )

        if mask:
            outmask = ak._v2.index.Index8.empty(outlength, self._nplike)
            assert outmask.nplike is self._nplike and parents.nplike is self._nplike
            self._handle_error(
                self._nplike[
                    "awkward_NumpyArray_reduce_mask_ByteMaskedArray_64",
                    outmask.dtype.type,
                    parents.dtype.type,
                ](
                    outmask.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )

            out = ak._v2.contents.ByteMaskedArray(
                outmask,
                out,
                False,
                None,
                None,
                self._nplike,
            )

        if keepdims:
            out = ak._v2.contents.RegularArray(
                out,
                1,
                self.length,
                None,
                None,
                self._nplike,
            )

        return out

    def _validityerror(self, path):
        if len(self.shape) == 0:
            return f'at {path} ("{type(self)}"): shape is zero-dimensional'
        for i, dim in enumerate(self.shape):
            if dim < 0:
                return f'at {path} ("{type(self)}"): shape[{i}] < 0'
        for i, stride in enumerate(self.strides):
            if stride % self.dtype.itemsize != 0:
                return f'at {path} ("{type(self)}"): shape[{i}] % itemsize != 0'
        return ""

    def _rpad(self, target, axis, depth, clip):
        if len(self.shape) == 0:
            raise ValueError("cannot rpad a scalar")
        elif len(self.shape) > 1 or not self.is_contiguous:
            return self.toRegularArray()._rpad(target, axis, depth, clip)
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis != depth:
            raise np.AxisError(f"axis={axis} exceeds the depth of this array({depth})")
        if not clip:
            if target < self.length:
                return self
            else:
                return self._rpad(target, posaxis, depth, clip=True)
        else:
            return self.rpad_axis0(target, clip=True)

    def _nbytes_part(self):
        result = self.data.nbytes
        if self.identifier is not None:
            result = result + self.identifier._nbytes_part()
        return result

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        if self._data.ndim != 1:
            return self.toRegularArray()._to_arrow(
                pyarrow, mask_node, validbytes, length, options
            )

        nparray = self.raw(numpy)
        storage_type = pyarrow.from_numpy_dtype(nparray.dtype)

        if issubclass(nparray.dtype.type, (bool, np.bool_)):
            nparray = ak._v2._connect.pyarrow.packbits(nparray)

        return pyarrow.Array.from_buffers(
            ak._v2._connect.pyarrow.to_awkwardarrow_type(
                storage_type, options["extensionarray"], mask_node, self
            ),
            length,
            [
                ak._v2._connect.pyarrow.to_validbits(validbytes),
                ak._v2._connect.pyarrow.to_length(nparray, length),
            ],
            null_count=ak._v2._connect.pyarrow.to_null_count(
                validbytes, options["count_nulls"]
            ),
        )

    def _to_numpy(self, allow_missing):
        out = self._nplike.asarray(self)
        if type(out).__module__.startswith("cupy."):
            return out.get()
        else:
            return out

    def _completely_flatten(self, nplike, options):
        return [self.raw(nplike).reshape(-1)]

    def _recursively_apply(
        self, action, depth, depth_context, lateral_context, options
    ):
        if self._data.ndim != 1 and options["numpy_to_regular"]:
            return self.toRegularArray()._recursively_apply(
                action, depth, depth_context, lateral_context, options
            )

        if options["return_array"]:

            def continuation():
                if options["keep_parameters"]:
                    return self
                else:
                    return NumpyArray(self._data, self._identifier, None, self._nplike)

        else:

            def continuation():
                pass

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def packed(self):
        return self.contiguous().toRegularArray()

    def _to_list(self, behavior):
        if self.parameter("__array__") == "byte":
            return ak._v2._util.tobytes(self._data)

        elif self.parameter("__array__") == "char":
            return ak._v2._util.tobytes(self._data).decode(errors="surrogateescape")

        else:
            out = self._to_list_custom(behavior)
            if out is not None:
                return out

            return self._data.tolist()

    def _to_nplike(self, nplike):
        return NumpyArray(
            self.raw(nplike),
            identifier=self.identifier,
            parameters=self.parameters,
            nplike=nplike,
        )

    def _to_json_custom(
        self,
        nan_string,
        infinity_string,
        minus_infinity_string,
        complex_real_string,
        complex_imag_string,
    ):
        cls = ak._v2._util.arrayclass(self, None)
        if cls.__getitem__ is not ak._v2.highlevel.Array.__getitem__:
            array = cls(self)
            out = [None] * self.length
            for i in range(self.length):
                out[i] = array[i]
            return out

    def _to_json(
        self,
        nan_string,
        infinity_string,
        minus_infinity_string,
        complex_real_string,
        complex_imag_string,
    ):
        if (
            self.parameter("__array__") == "byte"
            or self.parameter("__array__") == "char"
        ):
            return ak._v2._util.tobytes(self._data).decode(errors="surrogateescape")

        else:
            if self.dtype == np.complex128:
                if complex_real_string is None or complex_imag_string is None:
                    raise ValueError(
                        "Complex numbers can't be converted to JSON without"
                        " setting 'complex_record_fields' "
                    )

                return ak._v2.operations.structure.zip(
                    {
                        complex_real_string: ak._v2.contents.NumpyArray(
                            self._data.real
                        ),
                        complex_imag_string: ak._v2.contents.NumpyArray(
                            self._data.imag
                        ),
                    }
                ).layout._to_json(
                    nan_string,
                    infinity_string,
                    minus_infinity_string,
                    complex_real_string,
                    complex_imag_string,
                )

            if (
                nan_string is not None
                or infinity_string is not None
                or minus_infinity_string is not None
            ):
                out = self._nonfinite_to_union(
                    nan_string, infinity_string, minus_infinity_string
                )
                return out.tolist()

            out = self._to_json_custom(
                nan_string,
                infinity_string,
                minus_infinity_string,
                complex_real_string,
                complex_imag_string,
            )
            if out is not None:
                return out

            return self._data.tolist()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def num(array, axis=1, highlevel=True, behavior=None):
    """
    Args:
        array: Data containing nested lists to count.
        axis (int): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array of integers specifying the number of elements at a
    particular level.

    For instance, given the following doubly nested `array`,

        ak.Array([[
                   [1.1, 2.2, 3.3],
                   [],
                   [4.4, 5.5],
                   [6.6]
                  ],
                  [],
                  [
                   [7.7],
                   [8.8, 9.9]]
                  ])

    The number of elements in `axis=1` is

        >>> ak.num(array, axis=1)
        <Array [4, 0, 2] type='3 * int64'>

    and the number of elements at the next level down, `axis=2`, is

        >>> ak.num(array, axis=2)
        <Array [[3, 0, 2, 1], [], [1, 2]] type='3 * var * int64'>

    The `axis=0` case is special: it returns a scalar, the length of the array.

        >>> ak.num(array, axis=0)
        3

    This function is useful for ensuring that slices do not raise errors. For
    instance, suppose that we want to select the first element from each
    of the outermost nested lists of `array`. One of these lists is empty, so
    selecting the first element (`0`) would raise an error. However, if our
    first selection is `ak.num(array) > 0`, we are left with only those lists
    that *do* have a first element:

        >>> array[ak.num(array) > 0, 0]
        <Array [[1.1, 2.2, 3.3], [7.7]] type='2 * var * float64'>

    To keep a placeholder (None) in each place we do not want to select,
    consider using #ak.mask instead of a #ak.Array.__getitem__.

        >>> ak.mask(array, ak.num(array) > 0)[:, 0]
        <Array [[1.1, 2.2, 3.3], None, [7.7]] type='3 * option[var * float64]'>
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    out = layout.num(axis=axis)
    if isinstance(out, (ak._v2.contents.Content, ak._v2.record.Record)):
        return ak._v2._util.wrap(out, behavior, highlevel)
    else:
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def to_pandas(
    array, how="inner", levelname=lambda i: "sub" * i + "entry", anonymous="values"
):
    raise NotImplementedError


#     """
#     Args:
#         array: Data to convert into one or more Pandas DataFrames.
#         how (None or str): Passed to
#             [pd.merge](https://pandas.pydata.org/pandas-docs/version/1.0.3/reference/api/pandas.merge.html)
#             to combine DataFrames for each multiplicity into one DataFrame. If
#             None, a list of Pandas DataFrames is returned.
#         levelname (int -> str): Computes a name for each level of the row index
#             from the number of levels deep.
#         anonymous (str): Column name to use if the `array` does not contain
#             records; otherwise, column names are derived from record fields.

#     Converts Awkward data structures into Pandas
#     [MultiIndex](https://pandas.pydata.org/pandas-docs/stable/user_guide/advanced.html)
#     rows and columns. The resulting DataFrame(s) contains no Awkward structures.

#     #ak.Array structures can't be losslessly converted into a single DataFrame;
#     different fields in a record structure might have different nested list
#     lengths, but a DataFrame can have only one index.

#     If `how` is None, this function always returns a list of DataFrames (even
#     if it contains only one DataFrame); otherwise `how` is passed to
#     [pd.merge](https://pandas.pydata.org/pandas-docs/version/1.0.3/reference/api/pandas.merge.html)
#     to merge them into a single DataFrame with the associated loss of data.

#     In the following example, nested lists are converted into MultiIndex rows.
#     The index level names `"entry"`, `"subentry"` and `"subsubentry"` can be
#     controlled with the `levelname` parameter. The column name `"values"` is
#     assigned because this array has no fields; it can be controlled with the
#     `anonymous` parameter.

#         >>> ak.to_pandas(ak.Array([[[1.1, 2.2], [], [3.3]],
#         ...                        [],
#         ...                        [[4.4], [5.5, 6.6]],
#         ...                        [[7.7]],
#         ...                        [[8.8]]]))
#                                     values
#         entry subentry subsubentry
#         0     0        0               1.1
#                        1               2.2
#               2        0               3.3
#         2     0        0               4.4
#               1        0               5.5
#                        1               6.6
#         3     0        0               7.7
#         4     0        0               8.8

#     In this example, nested records are converted into MultiIndex columns.
#     (MultiIndex rows and columns can be mixed; these examples are deliberately
#     simple.)

#         >>> ak.to_pandas(ak.Array([
#         ...     {"I": {"a": _, "b": {"i": _}}, "II": {"x": {"y": {"z": _}}}}
#         ...     for _ in range(0, 50, 10)]))
#                 I      II
#                 a   b   x
#                     i   y
#                         z
#         entry
#         0       0   0   0
#         1      10  10  10
#         2      20  20  20
#         3      30  30  30
#         4      40  40  40

#     The following two examples show how fields of different length lists are
#     merged. With `how="inner"` (default), only subentries that exist for all
#     fields are preserved; with `how="outer"`, all subentries are preserved at
#     the expense of requiring missing values.

#         >>> ak.to_pandas(ak.Array([{"x": [], "y": [4.4, 3.3, 2.2, 1.1]},
#         ...                        {"x": [1], "y": [3.3, 2.2, 1.1]},
#         ...                        {"x": [1, 2], "y": [2.2, 1.1]},
#         ...                        {"x": [1, 2, 3], "y": [1.1]},
#         ...                        {"x": [1, 2, 3, 4], "y": []}]),
#         ...                        how="inner")
#                         x    y
#         entry subentry
#         1     0         1  3.3
#         2     0         1  2.2
#               1         2  1.1
#         3     0         1  1.1

#     The same with `how="outer"`:

#         >>> ak.to_pandas(ak.Array([{"x": [], "y": [4.4, 3.3, 2.2, 1.1]},
#         ...                        {"x": [1], "y": [3.3, 2.2, 1.1]},
#         ...                        {"x": [1, 2], "y": [2.2, 1.1]},
#         ...                        {"x": [1, 2, 3], "y": [1.1]},
#         ...                        {"x": [1, 2, 3, 4], "y": []}]),
#         ...                        how="outer")
#                           x    y
#         entry subentry
#         0     0         NaN  4.4
#               1         NaN  3.3
#               2         NaN  2.2
#               3         NaN  1.1
#         1     0         1.0  3.3
#               1         NaN  2.2
#               2         NaN  1.1
#         2     0         1.0  2.2
#               1         2.0  1.1
#         3     0         1.0  1.1
#               1         2.0  NaN
#               2         3.0  NaN
#         4     0         1.0  NaN
#               1         2.0  NaN
#               2         3.0  NaN
#               3         4.0  NaN
#     """
#     try:
#         import pandas
#     except ImportError:
#         raise ImportError(
#             """install the 'pandas' package with:

#     pip install pandas --upgrade

# or

#     conda install pandas"""
#         )

#     if how is not None:
#         out = None
#         for df in to_pandas(array, how=None, levelname=levelname, anonymous=anonymous):
#             if out is None:
#                 out = df
#             else:
#                 out = pandas.merge(out, df, how=how, left_index=True, right_index=True)
#         return out

#     def recurse(layout, row_arrays, col_names):
#         if isinstance(layout, ak._v2._util.virtualtypes):
#             return recurse(layout.array, row_arrays, col_names)

#         elif isinstance(layout, ak._v2._util.indexedtypes):
#             return recurse(layout.project(), row_arrays, col_names)

#         elif layout.parameter("__array__") in ("string", "bytestring"):
#             return [(to_numpy(layout), row_arrays, col_names)]

#         elif layout.purelist_depth > 1:
#             offsets, flattened = layout.offsets_and_flatten(axis=1)
#             offsets = numpy.asarray(offsets)
#             starts, stops = offsets[:-1], offsets[1:]
#             counts = stops - starts
#             if ak._v2._util.win or ak._v2._util.bits32:
#                 counts = counts.astype(np.int32)
#             if len(row_arrays) == 0:
#                 newrows = [
#                     numpy.repeat(numpy.arange(len(counts), dtype=counts.dtype), counts)
#                 ]
#             else:
#                 newrows = [numpy.repeat(x, counts) for x in row_arrays]
#             newrows.append(
#                 numpy.arange(offsets[-1], dtype=counts.dtype)
#                 - numpy.repeat(starts, counts)
#             )
#             return recurse(flattened, newrows, col_names)

#         elif isinstance(layout, ak._v2._util.uniontypes):
#             layout = ak._v2._util.union_to_record(layout, anonymous)
#             if isinstance(layout, ak._v2._util.uniontypes):
#                 return [(to_numpy(layout), row_arrays, col_names)]
#             else:
#                 return sum(
#                     [
#                         recurse(layout.field(n), row_arrays, col_names + (n,))
#                         for n in layout.keys()
#                     ],
#                     [],
#                 )

#         elif isinstance(layout, ak._v2.contents.RecordArray):
#             return sum(
#                 [
#                     recurse(layout.field(n), row_arrays, col_names + (n,))
#                     for n in layout.keys()
#                 ],
#                 [],
#             )

#         else:
#             return [(to_numpy(layout), row_arrays, col_names)]

#     layout = to_layout(array, allow_record=True, allow_other=False)
#     if isinstance(layout, ak.partition.PartitionedArray):   # NO PARTITIONED ARRAY
#         layout = layout.toContent()

#     if isinstance(layout, ak._v2.record.Record):
#         layout2 = layout.array[layout.at : layout.at + 1]
#     else:
#         layout2 = layout

#     tables = []
#     last_row_arrays = None
#     for column, row_arrays, col_names in recurse(layout2, [], ()):
#         if isinstance(layout, ak._v2.record.Record):
#             row_arrays = row_arrays[1:]  # Record --> one-element RecordArray
#         if len(col_names) == 0:
#             columns = [anonymous]
#         else:
#             columns = pandas.MultiIndex.from_tuples([col_names])

#         if (
#             last_row_arrays is not None
#             and len(last_row_arrays) == len(row_arrays)
#             and all(
#                 numpy.array_equal(x, y) for x, y in zip(last_row_arrays, row_arrays)
#             )
#         ):
#             oldcolumns = tables[-1].columns
#             if isinstance(oldcolumns, pandas.MultiIndex):
#                 numold = len(oldcolumns.levels)
#             else:
#                 numold = max(len(x) for x in oldcolumns)
#             numnew = len(columns.levels)
#             maxnum = max(numold, numnew)
#             if numold != maxnum:
#                 oldcolumns = pandas.MultiIndex.from_tuples(
#                     [x + ("",) * (maxnum - numold) for x in oldcolumns]
#                 )
#                 tables[-1].columns = oldcolumns
#             if numnew != maxnum:
#                 columns = pandas.MultiIndex.from_tuples(
#                     [x + ("",) * (maxnum - numnew) for x in columns]
#                 )

#             newframe = pandas.DataFrame(
#                 data=column, index=tables[-1].index, columns=columns
#             )
#             tables[-1] = pandas.concat([tables[-1], newframe], axis=1)

#         else:
#             if len(row_arrays) == 0:
#                 index = pandas.RangeIndex(len(column), name=levelname(0))
#             else:
#                 index = pandas.MultiIndex.from_arrays(
#                     row_arrays, names=[levelname(i) for i in range(len(row_arrays))]
#                 )
#             tables.append(pandas.DataFrame(data=column, index=index, columns=columns))

#         last_row_arrays = row_arrays

#     for table in tables:
#         if isinstance(table.columns, pandas.MultiIndex) and table.columns.nlevels == 1:
#             table.columns = table.columns.get_level_values(0)

#     return tables



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from awkward._v2.operations.describe.ak_is_valid import is_valid  # noqa: F401
from awkward._v2.operations.describe.ak_validity_error import (  # noqa: F401
    validity_error,
)
from awkward._v2.operations.describe.ak_type import type  # noqa: F401
from awkward._v2.operations.describe.ak_parameters import parameters  # noqa: F401
from awkward._v2.operations.describe.ak_fields import fields  # noqa: F401
from awkward._v2.operations.describe.ak_backend import backend  # noqa: F401
from awkward._v2.operations.describe.ak_to_backend import to_backend  # noqa: F401



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


class Reducer:
    needs_position = False

    @classmethod
    def return_dtype(cls, given_dtype):
        if given_dtype in (np.bool_, np.int8, np.int16, np.int32):
            return np.int32 if ak._util.win or ak._util.bits32 else np.int64

        if given_dtype in (np.uint8, np.uint16, np.uint32):
            return np.uint32 if ak._util.win or ak._util.bits32 else np.uint64

        if given_dtype == np.complex128:
            return np.float64
        if given_dtype == np.complex64:
            return np.float32

        return given_dtype

    @classmethod
    def maybe_double_length(cls, type, length):
        return 2 * length if type in (np.complex128, np.complex64) else length

    @classmethod
    def maybe_other_type(cls, dtype):
        type = np.int64 if dtype.kind.upper() == "M" else dtype.type
        if dtype == np.complex128:
            type = np.float64
        if dtype == np.complex64:
            type = np.float32
        return type


class ArgMin(Reducer):
    name = "argmin"
    needs_position = True
    preferred_dtype = np.int64

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        dtype = cls.maybe_other_type(array.dtype)
        result = array.nplike.empty(outlength, dtype=np.int64)
        if array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_argmin_complex",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_argmin",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        return ak._v2.contents.NumpyArray(result)


class ArgMax(Reducer):
    name = "argmax"
    needs_position = True
    preferred_dtype = np.int64

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        dtype = cls.maybe_other_type(array.dtype)
        result = array.nplike.empty(outlength, dtype=np.int64)
        if array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_argmax_complex",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_argmax",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        return ak._v2.contents.NumpyArray(result)


class Count(Reducer):
    name = "count"
    preferred_dtype = np.float64

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        result = array.nplike.empty(outlength, dtype=np.int64)
        assert parents.nplike is array.nplike
        array._handle_error(
            array.nplike[
                "awkward_reduce_count_64", result.dtype.type, parents.dtype.type
            ](
                result,
                parents.data,
                parents.length,
                outlength,
            )
        )
        return ak._v2.contents.NumpyArray(result)


class CountNonzero(Reducer):
    name = "count_nonzero"
    preferred_dtype = np.float64

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        dtype = np.dtype(np.int64) if array.dtype.kind.upper() == "M" else array.dtype
        result = array.nplike.empty(outlength, dtype=np.int64)
        if array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_countnonzero_complex",
                    result.dtype.type,
                    cls.return_dtype(array.dtype),
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_countnonzero",
                    result.dtype.type,
                    dtype.type,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        return ak._v2.contents.NumpyArray(result)


class Sum(Reducer):
    name = "sum"
    preferred_dtype = np.float64

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        if array.dtype.kind == "M":
            raise ValueError(f"cannot compute the sum (ak.sum) of {array.dtype!r}")
        else:
            dtype = cls.maybe_other_type(array.dtype)
        result = array.nplike.empty(
            cls.maybe_double_length(array.dtype.type, outlength),
            dtype=cls.return_dtype(dtype),
        )

        if array.dtype == np.bool_:
            if result.dtype in (np.int64, np.uint64):
                assert parents.nplike is array.nplike
                assert parents.nplike is array.nplike
                array._handle_error(
                    array.nplike[
                        "awkward_reduce_sum_int64_bool_64",
                        np.int64,
                        array.dtype.type,
                        parents.dtype.type,
                    ](
                        result,
                        array.data,
                        parents.data,
                        parents.length,
                        outlength,
                    )
                )
            elif result.dtype in (np.int32, np.uint32):
                assert parents.nplike is array.nplike
                array._handle_error(
                    array.nplike[
                        "awkward_reduce_sum_int32_bool_64",
                        np.int32,
                        array.dtype.type,
                        parents.dtype.type,
                    ](
                        result,
                        array.data,
                        parents.data,
                        parents.length,
                        outlength,
                    )
                )
            else:
                raise NotImplementedError
        elif array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_sum_complex",
                    result.dtype.type,
                    np.float64 if array.dtype.type == np.complex128 else np.float32,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_sum",
                    result.dtype.type,
                    np.int64 if array.dtype.kind == "m" else array.dtype.type,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )

        if array.dtype.kind == "m":
            return ak._v2.contents.NumpyArray(array.nplike.asarray(result, array.dtype))
        elif array.dtype.type in (np.complex128, np.complex64):
            return ak._v2.contents.NumpyArray(result.view(array.dtype))
        else:
            return ak._v2.contents.NumpyArray(result)


class Prod(Reducer):
    name = "prod"
    preferred_dtype = np.int64

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        if array.dtype.kind.upper() == "M":
            raise ValueError(f"cannot compute the product (ak.prod) of {array.dtype!r}")
        result = array.nplike.empty(
            cls.maybe_double_length(array.dtype.type, outlength),
            dtype=cls.return_dtype(array.dtype),
        )
        if array.dtype == np.bool_:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod_bool",
                    array.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        elif array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod_complex",
                    result.dtype.type,
                    np.float64 if array.dtype.type == np.complex128 else np.float32,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        if array.dtype.type in (np.complex128, np.complex64):
            return ak._v2.contents.NumpyArray(result.view(array.dtype))
        else:
            return ak._v2.contents.NumpyArray(result)


class Any(Reducer):
    name = "any"
    preferred_dtype = np.bool_

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        dtype = cls.maybe_other_type(array.dtype)
        result = array.nplike.empty(outlength, dtype=np.bool_)
        if array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_sum_bool_complex",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_sum_bool",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        return ak._v2.contents.NumpyArray(result)


class All(Reducer):
    name = "all"
    preferred_dtype = np.bool_

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        dtype = cls.maybe_other_type(array.dtype)
        result = array.nplike.empty(outlength, dtype=np.bool_)
        if array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod_bool_complex",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod_bool",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        return ak._v2.contents.NumpyArray(result)


class Min(Reducer):
    name = "min"
    preferred_dtype = np.float64
    initial = None

    def __init__(self, initial):
        type(self).initial = initial

    def __del__(self):
        type(self).initial = None

    @staticmethod
    def _min_initial(initial, type):
        if initial is None:
            if type in (
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ):
                return np.iinfo(type).max
            else:
                return np.inf

        return initial

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        dtype = cls.maybe_other_type(array.dtype)
        result = array.nplike.empty(
            cls.maybe_double_length(array.dtype.type, outlength), dtype=dtype
        )
        if array.dtype == np.bool_:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_prod_bool",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        elif array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_min_complex",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                    cls._min_initial(cls.initial, dtype),
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_min",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                    cls._min_initial(cls.initial, dtype),
                )
            )
        if array.dtype.type in (np.complex128, np.complex64):
            return ak._v2.contents.NumpyArray(
                array.nplike.array(result.view(array.dtype), array.dtype)
            )
        else:
            return ak._v2.contents.NumpyArray(array.nplike.array(result, array.dtype))


class Max(Reducer):
    name = "max"
    preferred_dtype = np.float64
    initial = None

    def __init__(self, initial):
        type(self).initial = initial

    def __del__(self):
        type(self).initial = None

    @staticmethod
    def _max_initial(initial, type):
        if initial is None:
            if type in (
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ):
                return np.iinfo(type).min
            else:
                return -np.inf

        return initial

    @classmethod
    def apply(cls, array, parents, outlength):
        assert isinstance(array, ak._v2.contents.NumpyArray)
        dtype = cls.maybe_other_type(array.dtype)
        result = array.nplike.empty(
            cls.maybe_double_length(array.dtype.type, outlength), dtype=dtype
        )
        if array.dtype == np.bool_:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_sum_bool",
                    result.dtype.type,
                    array.dtype.type,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )
        elif array.dtype.type in (np.complex128, np.complex64):
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_max_complex",
                    dtype,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                    cls._max_initial(cls.initial, dtype),
                )
            )
        else:
            assert parents.nplike is array.nplike
            array._handle_error(
                array.nplike[
                    "awkward_reduce_max",
                    result.dtype.type,
                    dtype,
                    parents.dtype.type,
                ](
                    result,
                    array.data,
                    parents.data,
                    parents.length,
                    outlength,
                    cls._max_initial(cls.initial, dtype),
                )
            )
        if array.dtype.type in (np.complex128, np.complex64):
            return ak._v2.contents.NumpyArray(
                array.nplike.array(result.view(array.dtype), array.dtype)
            )
        else:
            return ak._v2.contents.NumpyArray(array.nplike.array(result, array.dtype))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401


def test_import_and_loading_shared_lib():
    try:
        import awkward_cuda_kernels
    except ModuleNotFoundError:
        pytest.fail("Could not import awkward_cuda_kernels")

    try:
        import ctypes

        cuak = ctypes.cdll.LoadLibrary(awkward_cuda_kernels.shared_library_path)

        if not hasattr(cuak, "awkward_cuda_ptr_device_name"):
            pytest.fail("Cannot access the functions in the loaded shared library")
    except Exception:
        pytest.fail("Could not load the shared library or access the function")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def tt(highlevel):
    return ak._v2.highlevel.Array(highlevel.layout.typetracer)


def test_basic():
    array = ak._v2.highlevel.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    assert to_list(array + array) == [[2.2, 4.4, 6.6], [], [8.8, 11.0]]
    assert (array + array).layout.form == (tt(array) + tt(array)).layout.form
    assert to_list(array * 2) == [[2.2, 4.4, 6.6], [], [8.8, 11.0]]
    assert to_list(2 * array) == [[2.2, 4.4, 6.6], [], [8.8, 11.0]]
    assert (array * 2).layout.form == (tt(array) * 2).layout.form
    assert (array * 2).layout.form == (2 * tt(array)).layout.form


def test_emptyarray():
    one = ak._v2.highlevel.Array(ak._v2.contents.NumpyArray(np.array([])))
    two = ak._v2.highlevel.Array(ak._v2.contents.EmptyArray())
    assert to_list(one + one) == []
    assert to_list(two + two) == []
    assert to_list(one + two) == []
    assert (one + one).layout.form == (tt(one) + tt(one)).layout.form
    assert (two + two).layout.form == (tt(two) + tt(two)).layout.form
    assert (one + two).layout.form == (tt(one) + tt(two)).layout.form


def test_indexedarray():
    content = ak._v2.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index1 = ak._v2.index.Index64(np.array([2, 4, 4, 0, 8], dtype=np.int64))
    index2 = ak._v2.index.Index64(np.array([6, 4, 4, 8, 0], dtype=np.int64))
    one = ak._v2.highlevel.Array(ak._v2.contents.IndexedArray(index1, content))
    two = ak._v2.highlevel.Array(ak._v2.contents.IndexedArray(index2, content))
    assert to_list(one + two) == [8.8, 8.8, 8.8, 8.8, 8.8]
    assert (one + two).layout.form == (tt(one) + tt(two)).layout.form


def test_indexedoptionarray():
    content = ak._v2.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index1 = ak._v2.index.Index64(np.array([2, -1, 4, 0, 8], dtype=np.int64))
    index2 = ak._v2.index.Index64(np.array([-1, 4, 4, -1, 0], dtype=np.int64))
    one = ak._v2.highlevel.Array(ak._v2.contents.IndexedOptionArray(index1, content))
    two = ak._v2.highlevel.Array(ak._v2.contents.IndexedOptionArray(index2, content))
    assert to_list(one + two) == [None, None, 8.8, None, 8.8]
    assert (one + two).layout.form == (tt(one) + tt(two)).layout.form

    uno = ak._v2.highlevel.Array(
        ak._v2.contents.NumpyArray(np.array([2.2, 4.4, 4.4, 0.0, 8.8]))
    )
    dos = ak._v2.highlevel.Array(
        ak._v2.contents.NumpyArray(np.array([6.6, 4.4, 4.4, 8.8, 0.0]))
    )
    assert to_list(uno + two) == [None, 8.8, 8.8, None, 8.8]
    assert (uno + two).layout.form == (tt(uno) + tt(two)).layout.form
    assert to_list(one + dos) == [8.8, None, 8.8, 8.8, 8.8]
    assert (one + dos).layout.form == (tt(one) + tt(dos)).layout.form


def test_regularize_shape():
    array = ak._v2.contents.NumpyArray(np.arange(2 * 3 * 5).reshape(2, 3, 5))
    assert isinstance(array.toRegularArray(), ak._v2.contents.RegularArray)
    assert to_list(array.toRegularArray()) == to_list(array)


def test_regulararray():
    array = ak._v2.highlevel.Array(np.arange(2 * 3 * 5).reshape(2, 3, 5))
    assert (
        to_list(array + array) == (np.arange(2 * 3 * 5).reshape(2, 3, 5) * 2).tolist()
    )
    assert (array + array).layout.form == (tt(array) + tt(array)).layout.form
    assert to_list(array * 2) == (np.arange(2 * 3 * 5).reshape(2, 3, 5) * 2).tolist()
    assert (array * 2).layout.form == (tt(array) * 2).layout.form
    array2 = ak._v2.highlevel.Array(np.arange(2 * 1 * 5).reshape(2, 1, 5))
    assert to_list(array + array2) == to_list(
        np.arange(2 * 3 * 5).reshape(2, 3, 5) + np.arange(2 * 1 * 5).reshape(2, 1, 5)
    )
    assert (array + array2).layout.form == (tt(array) + tt(array2)).layout.form
    array3 = ak._v2.highlevel.Array(np.arange(2 * 3 * 5).reshape(2, 3, 5).tolist())
    assert to_list(array + array3) == to_list(
        np.arange(2 * 3 * 5).reshape(2, 3, 5) + np.arange(2 * 3 * 5).reshape(2, 3, 5)
    )
    assert (array + array3).layout.form == (tt(array) + tt(array3)).layout.form
    assert to_list(array3 + array) == to_list(
        np.arange(2 * 3 * 5).reshape(2, 3, 5) + np.arange(2 * 3 * 5).reshape(2, 3, 5)
    )
    assert (array3 + array).layout.form == (tt(array3) + tt(array)).layout.form


def test_listarray():
    content = ak._v2.contents.NumpyArray(np.arange(12, dtype=np.int64))
    starts = ak._v2.index.Index64(np.array([3, 0, 999, 2, 6, 10], dtype=np.int64))
    stops = ak._v2.index.Index64(np.array([7, 3, 999, 4, 6, 12], dtype=np.int64))
    one = ak._v2.highlevel.Array(ak._v2.contents.ListArray(starts, stops, content))
    two = ak._v2.highlevel.Array(
        [[100, 100, 100, 100], [200, 200, 200], [], [300, 300], [], [400, 400]]
    )
    assert to_list(one) == [[3, 4, 5, 6], [0, 1, 2], [], [2, 3], [], [10, 11]]
    assert to_list(one + 100) == [
        [103, 104, 105, 106],
        [100, 101, 102],
        [],
        [102, 103],
        [],
        [110, 111],
    ]
    assert (one + 100).layout.form == (tt(one) + 100).layout.form
    assert to_list(one + two) == [
        [103, 104, 105, 106],
        [200, 201, 202],
        [],
        [302, 303],
        [],
        [410, 411],
    ]
    assert (one + two).layout.form == (tt(one) + tt(two)).layout.form
    assert to_list(two + one) == [
        [103, 104, 105, 106],
        [200, 201, 202],
        [],
        [302, 303],
        [],
        [410, 411],
    ]
    assert (two + one).layout.form == (tt(two) + tt(one)).layout.form
    assert to_list(one + np.array([100, 200, 300, 400, 500, 600])[:, np.newaxis]) == [
        [103, 104, 105, 106],
        [200, 201, 202],
        [],
        [402, 403],
        [],
        [610, 611],
    ]
    assert to_list(np.array([100, 200, 300, 400, 500, 600])[:, np.newaxis] + one) == [
        [103, 104, 105, 106],
        [200, 201, 202],
        [],
        [402, 403],
        [],
        [610, 611],
    ]
    assert to_list(one + 100) == [
        [103, 104, 105, 106],
        [100, 101, 102],
        [],
        [102, 103],
        [],
        [110, 111],
    ]
    assert (one + 100).layout.form == (tt(one) + 100).layout.form


def test_unionarray():
    one0 = ak._v2.contents.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64))
    one1 = ak._v2.contents.NumpyArray(np.array([4, 5], dtype=np.int64))
    onetags = ak._v2.index.Index8(np.array([0, 0, 0, 0, 1, 1], dtype=np.int8))
    oneindex = ak._v2.index.Index64(np.array([0, 1, 2, 3, 0, 1], dtype=np.int64))
    one = ak._v2.highlevel.Array(
        ak._v2.contents.UnionArray(onetags, oneindex, [one0, one1])
    )

    two0 = ak._v2.contents.NumpyArray(np.array([0, 100], dtype=np.int64))
    two1 = ak._v2.contents.NumpyArray(
        np.array([200.3, 300.3, 400.4, 500.5], dtype=np.float64)
    )
    twotags = ak._v2.index.Index8(np.array([0, 0, 1, 1, 1, 1], dtype=np.int8))
    twoindex = ak._v2.index.Index64(np.array([0, 1, 0, 1, 2, 3], dtype=np.int64))
    two = ak._v2.highlevel.Array(
        ak._v2.contents.UnionArray(twotags, twoindex, [two0, two1])
    )

    assert to_list(one) == [0.0, 1.1, 2.2, 3.3, 4, 5]
    assert to_list(two) == [0, 100, 200.3, 300.3, 400.4, 500.5]
    assert to_list(one + two) == [0.0, 101.1, 202.5, 303.6, 404.4, 505.5]
    assert to_list(one + 100) == [100.0, 101.1, 102.2, 103.3, 104, 105]
    assert to_list(100 + one) == [100.0, 101.1, 102.2, 103.3, 104, 105]
    assert (one + two).layout.form == (tt(one) + tt(two)).layout.form
    assert (one + 100).layout.form == (tt(one) + 100).layout.form
    assert (100 + one).layout.form == (100 + tt(one)).layout.form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
from awkward._v2.highlevel import Array

np = ak.nplike.NumpyMetadata.instance()


class ByteBehavior(Array):
    __name__ = "Array"

    def __bytes__(self):
        tmp = ak.nplike.of(self.layout).asarray(self.layout)
        if hasattr(tmp, "tobytes"):
            return tmp.tobytes()
        else:
            return tmp.tostring()

    def __str__(self):
        return str(self.__bytes__())


#     def __repr__(self):
#         return repr(self.__bytes__())

#     def __iter__(self):
#         for x in self.__bytes__():
#             yield x

#     def __eq__(self, other):
#         if isinstance(other, (bytes, ByteBehavior)):
#             return bytes(self) == bytes(other)
#         else:
#             return False

#     def __ne__(self, other):
#         return not self.__eq__(self, other)

#     def __add__(self, other):
#         if isinstance(other, (bytes, ByteBehavior)):
#             return bytes(self) + bytes(other)
#         else:
#             raise TypeError("can only concatenate bytes to bytes")

#     def __radd__(self, other):
#         if isinstance(other, (bytes, ByteBehavior)):
#             return bytes(other) + bytes(self)
#         else:
#             raise TypeError("can only concatenate bytes to bytes")


class CharBehavior(Array):
    __name__ = "Array"

    def __bytes__(self):
        tmp = ak.nplike.of(self.layout).asarray(self.layout)
        if hasattr(tmp, "tobytes"):
            return tmp.tobytes()
        else:
            return tmp.tostring()

    def __str__(self):
        return self.__bytes__().decode("utf-8", "surrogateescape")


#     def __repr__(self):
#         return repr(self.__bytes__().decode("utf-8", "surrogateescape"))

#     def __iter__(self):
#         for x in self.__str__():
#             yield x

#     def __eq__(self, other):
#         if isinstance(other, (str, CharBehavior)):
#             return str(self) == str(other)
#         else:
#             return False

#     def __ne__(self, other):
#         return not self.__eq__(other)

#     def __add__(self, other):
#         if isinstance(other, (str, CharBehavior)):
#             return str(self) + str(other)
#         else:
#             raise TypeError("can only concatenate str to str")

#     def __radd__(self, other):
#         if isinstance(other, (str, CharBehavior)):
#             return str(other) + str(self)
#         else:
#             raise TypeError("can only concatenate str to str")


# class ByteStringBehavior(ak._v2.highlevel.Array):
#     __name__ = "Array"

#     def __iter__(self):
#         for x in super(ByteStringBehavior, self).__iter__():
#             yield x.__bytes__()


# class StringBehavior(ak._v2.highlevel.Array):
#     __name__ = "Array"

#     def __iter__(self):
#         for x in super(StringBehavior, self).__iter__():
#             yield x.__str__()


# def _string_equal(one, two):
#     nplike = ak.nplike.of(one, two)
#     behavior = ak._v2._util.behaviorof(one, two)

#     one, two = ak.without_parameters(one).layout, ak.without_parameters(two).layout

#     # first condition: string lengths must be the same
#     counts1 = nplike.asarray(one.count(axis=-1))
#     counts2 = nplike.asarray(two.count(axis=-1))

#     out = counts1 == counts2

#     # only compare characters in strings that are possibly equal (same length)
#     possible = nplike.logical_and(out, counts1)
#     possible_counts = counts1[possible]

#     if len(possible_counts) > 0:
#         onepossible = one[possible]
#         twopossible = two[possible]

#         reduced = ak.all(ak.Array(onepossible) == ak.Array(twopossible), axis=-1).layout

#         # update same-length strings with a verdict about their characters
#         out[possible] = reduced

#     return ak._v2._util.wrap(ak._v2.contents.NumpyArray(out), behavior)


# def _string_notequal(one, two):
#     return ~_string_equal(one, two)


# def _string_broadcast(layout, offsets):
#     nplike = ak.nplike.of(offsets)
#     offsets = nplike.asarray(offsets)
#     counts = offsets[1:] - offsets[:-1]
#     if ak._v2._util.win or ak._v2._util.bits32:
#         counts = counts.astype(np.int32)
#     parents = nplike.repeat(nplike.arange(len(counts), dtype=counts.dtype), counts)
#     return ak._v2.contents.IndexedArray64(ak._v2.index.Index64(parents), layout).project()


def _string_numba_typer(viewtype):
    import numba

    if viewtype.type.parameters["__array__"] == "string":
        return numba.types.string
    else:
        return numba.cpython.charseq.bytes_type


def _string_numba_lower(
    context, builder, rettype, viewtype, viewval, viewproxy, attype, atval
):
    import numba
    import llvmlite.llvmpy.core

    whichpos = ak._v2._connect.numba.layout.posat(
        context, builder, viewproxy.pos, viewtype.type.CONTENT
    )
    nextpos = ak._v2._connect.numba.layout.getat(
        context, builder, viewproxy.arrayptrs, whichpos
    )

    whichnextpos = ak._v2._connect.numba.layout.posat(
        context, builder, nextpos, viewtype.type.contenttype.ARRAY
    )

    startspos = ak._v2._connect.numba.layout.posat(
        context, builder, viewproxy.pos, viewtype.type.STARTS
    )
    startsptr = ak._v2._connect.numba.layout.getat(
        context, builder, viewproxy.arrayptrs, startspos
    )
    startsarraypos = builder.add(viewproxy.start, atval)
    start = ak._v2._connect.numba.layout.getat(
        context, builder, startsptr, startsarraypos, viewtype.type.indextype.dtype
    )

    stopspos = ak._v2._connect.numba.layout.posat(
        context, builder, viewproxy.pos, viewtype.type.STOPS
    )
    stopsptr = ak._v2._connect.numba.layout.getat(
        context, builder, viewproxy.arrayptrs, stopspos
    )
    stopsarraypos = builder.add(viewproxy.start, atval)
    stop = ak._v2._connect.numba.layout.getat(
        context, builder, stopsptr, stopsarraypos, viewtype.type.indextype.dtype
    )

    baseptr = ak._v2._connect.numba.layout.getat(
        context, builder, viewproxy.arrayptrs, whichnextpos
    )
    rawptr = builder.add(
        baseptr,
        ak._v2._connect.numba.layout.castint(
            context, builder, viewtype.type.indextype.dtype, numba.intp, start
        ),
    )
    rawptr_cast = builder.inttoptr(
        rawptr,
        llvmlite.llvmpy.core.Type.pointer(
            llvmlite.llvmpy.core.Type.int(numba.intp.bitwidth // 8)
        ),
    )
    strsize = builder.sub(stop, start)
    strsize_cast = ak._v2._connect.numba.layout.castint(
        context, builder, viewtype.type.indextype.dtype, numba.intp, strsize
    )

    pyapi = context.get_python_api(builder)
    gil = pyapi.gil_ensure()

    strptr = builder.bitcast(rawptr_cast, pyapi.cstring)

    if viewtype.type.parameters["__array__"] == "string":
        kind = context.get_constant(numba.types.int32, pyapi.py_unicode_1byte_kind)
        pystr = pyapi.string_from_kind_and_data(kind, strptr, strsize_cast)
    else:
        pystr = pyapi.bytes_from_string_and_size(strptr, strsize_cast)

    out = pyapi.to_native_value(rettype, pystr).value

    pyapi.gil_release(gil)

    return out


def register(behavior):
    # behavior["byte"] = ByteBehavior
    # behavior["__typestr__", "byte"] = "byte"
    # behavior["char"] = CharBehavior
    # behavior["__typestr__", "char"] = "char"

    # behavior["bytestring"] = ByteStringBehavior
    # behavior["__typestr__", "bytestring"] = "bytes"
    # behavior["string"] = StringBehavior
    # behavior["__typestr__", "string"] = "string"

    # behavior[ak.nplike.numpy.equal, "bytestring", "bytestring"] = _string_equal
    # behavior[ak.nplike.numpy.equal, "string", "string"] = _string_equal
    # behavior[ak.nplike.numpy.not_equal, "bytestring", "bytestring"] = _string_notequal
    # behavior[ak.nplike.numpy.not_equal, "string", "string"] = _string_notequal

    # behavior["__broadcast__", "bytestring"] = _string_broadcast
    # behavior["__broadcast__", "string"] = _string_broadcast

    behavior["__numba_typer__", "bytestring"] = _string_numba_typer
    behavior["__numba_lower__", "bytestring"] = _string_numba_lower
    behavior["__numba_typer__", "string"] = _string_numba_typer
    behavior["__numba_lower__", "string"] = _string_numba_lower



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401
from awkward._v2.tmp_for_testing import v1_to_v2


def test():
    array = ak.Array(
        [[{"x": [[1, 2], []]}, {"x": []}, {"x": [[3]]}]], with_name="Record"
    )
    array_v2 = ak._v2.highlevel.Array(v1_to_v2(array.layout))

    # Flattening inside record does not drop parameters
    assert "__record__" not in ak.parameters(ak.flatten(array, axis=3))
    assert "__record__" not in array_v2.layout.flatten(axis=3).parameters

    # Flattening before record does not drop parameters
    assert "__record__" in ak.parameters(ak.flatten(array, axis=1))
    assert "__record__" in array_v2.layout.flatten(axis=1).parameters



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def from_jax(array, regulararray=False, highlevel=True, behavior=None):
    raise NotImplementedError


#     """
#     Args:
#         array (jax.numpy.array): The `jax.numpy.array` to convert into an Awkward Array.
#         regulararray (bool): If True and the array is multidimensional,
#             the dimensions are represented by nested #ak.layout.RegularArray
#             nodes; if False and the array is multidimensional, the dimensions
#             are represented by a multivalued #ak.layout.NumpyArray.shape.
#             If the array is one-dimensional, this has no effect.
#         highlevel (bool): If True, return an #ak.Array; otherwise, return
#             a low-level #ak.layout.Content subclass.
#         behavior (None or dict): Custom #ak.behavior for the output array, if
#             high-level.

#     Converts a JAX array into an Awkward Array.

#     The resulting layout may involve the following #ak.layout.Content types
#     (only):

#        * #ak.layout.NumpyArray
#        * #ak.layout.RegularArray if `regulararray=True`.

#     See also #ak.from_cupy and #ak.from_numpy.
#     """

#     def recurse(array):
#         if regulararray and len(array.shape) > 1:
#             return ak._v2.contents.RegularArray(
#                 recurse(array.reshape((-1,) + array.shape[2:])),
#                 array.shape[1],
#                 array.shape[0],
#             )

#         if len(array.shape) == 0:
#             data = ak._v2.contents.NumpyArray.from_jax(array.reshape(1))
#         else:
#             data = ak._v2.contents.NumpyArray.from_jax(array)

#         return data

#     layout = recurse(array)

#     return ak._v2._util.maybe_wrap(layout, behavior, highlevel)
